/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  f2_f2.cpp
///  \brief Implements the arithmetic in F_2[X] / (X^2)
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef F2_F2_HPP_
#define F2_F2_HPP_

#include "ring.hpp"
#include "ringelement.hpp"
#include "weights.hpp"

class F2_F2 : public Ring
{
	// 2 = X, 3 = 1+X
public:
    typedef SymWeight<2> MyWeight;
    static std::string myName(void) {return "F2[X] / (X^2)";}
    static unsigned int getSpacing(void) { return 2; };

	inline static Element<F2_F2> add(const Element<F2_F2> & a, const Element<F2_F2> & b) ;
	inline static Element<F2_F2> addInverse(const Element<F2_F2> &  a);
	inline static Element<F2_F2> minus(const Element<F2_F2> &  a, const Element<F2_F2> & b);
	inline static std::ostream& print(std::ostream& os, const Element<F2_F2> & a);
	inline static std::vector<Element<F2_F2> > multGenerator(void);

	// call the corresponding function from Ring
    inline static bool isUnit(const Element<F2_F2>& a);
    inline static Element<F2_F2> mult(const Element<F2_F2>& a, const Element<F2_F2>& b);
    inline static Element<F2_F2> multInverse(const Element<F2_F2>& a); // returns 0 if nilpotent
    inline static Element<F2_F2> div(const Element<F2_F2>& a, const Element<F2_F2>& b);
    inline static unsigned int height(const Element<F2_F2>& a) ;
    inline static bool isReducable(const Element<F2_F2>& a);
    inline static int compare(const Element<F2_F2>& a, const Element<F2_F2>& b);
};

Element<F2_F2> F2_F2::add(const Element<F2_F2> &  a, const Element<F2_F2>& b)
{
	switch(a.getIndex())
	{
	case 0: return b;
	case 1: 
		switch(b.getIndex())
		{
		case 0: return 1;
		case 1: return 0;
		case 2: return 3;
		case 3: return 2;
		default: std::cout<<"Should not happen in F2_F2::add"; exit(-1);
		}
	case 2:
		switch(b.getIndex())
		{
		case 0: return 2;
		case 1: return 3;
		case 2: return 0;
		case 3: return 1;
		default: std::cout<<"Should not happen in F2_F2::add"; exit(-1);
		}
	case 3:
		switch(b.getIndex())
		{
		case 0: return 3;
		case 1: return 2;
		case 2: return 1;
		case 3: return 0;
		default: std::cout<<"Should not happen in F2_F2::add"; exit(-1);
		}
	default: std::cout<<"Should not happen in F2_F2::add"; exit(-1);
	}
	std::cout<<"Should not happen in F2_F2::add"; 
	exit(-1);
}

Element<F2_F2> F2_F2::addInverse(const Element<F2_F2>& a)
{
	return a;
}

Element<F2_F2> F2_F2::minus(const Element<F2_F2>&  a, const Element<F2_F2>& b)
{
	return add(a,b);
}

std::ostream& F2_F2::print(std::ostream& os, const Element<F2_F2>& a)
{
    return os<<a.getIndex();

	switch(a.getIndex())
	{
	case 2: 
		os<<"X";
		break;
	case 3: 
		os<<"1+X";
		break;
	default:
		os<<(int) a.getIndex();
	}
	return os;
}

std::vector<Element<F2_F2> > F2_F2::multGenerator(void)
{
    return std::vector<Element<F2_F2> >(1,3);
}


bool F2_F2::isUnit(const Element<F2_F2>& a)
{
    return Ring::isUnit(a.getIndex());
}
Element<F2_F2> F2_F2::mult(const Element<F2_F2>& a, const Element<F2_F2>& b)
{
    return Ring::mult(a.getIndex(), b.getIndex());
}

Element<F2_F2> F2_F2::multInverse(const Element<F2_F2>& a)
{
    return Ring::multInverse(a.getIndex());
}

Element<F2_F2> F2_F2::div(const Element<F2_F2>& a, const Element<F2_F2>& b)
{
    return Ring::div(a.getIndex(), b.getIndex());
}

unsigned int F2_F2::height(const Element<F2_F2>& a)
{
    return Ring::height(a.getIndex());
}

bool F2_F2::isReducable(const Element<F2_F2>& a)
{
    return Ring::isReducable(a.getIndex());
}

int F2_F2::compare(const Element<F2_F2>& a, const Element<F2_F2>& b)
{
    return Ring::compare(a.getIndex(), b.getIndex());
}

#endif /*F2_F2_HPP_*/
