/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  field.hpp
///  \brief Implementation of finite fields.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   01.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef FIELD_HPP_
#define FIELD_HPP_

#include <cstdlib>
#include <vector>
#include <sstream>
#include "weights.hpp"
#include "ringelement.hpp"

class Field
{
private:
    static unsigned int m_p; ///< The characteristic of the field.
    static unsigned int m_r; ///< The order of the Galois group.
    static unsigned int m_q; ///< The order of the field.
    static unsigned int m_bitsPerEntry; ///< how many bits are needed to represent on entry in the field


    static INDEX m_alpha; ///< A primitive Element of the field.

    static std::vector<INDEX> m_primeFieldElements; ///< The Elements of the prime Field
    static std::vector<INDEX> m_addition; ///< The addition table.
    static std::vector<INDEX> m_subtraction; ///< The subtraction  table.
    static std::vector<INDEX> m_multiplication; ///< The multiplication table.
    static std::vector<INDEX> m_division; ///< The multiplication table.
    static std::vector<INDEX> m_additive_inv; ///< The vector of additive Inverse.
    static std::vector<INDEX> m_multiplicative_inv; ///< The vector of multiplicative Inverse.
    static std::vector<std::vector<INDEX> > m_frobenius; ///< The images of the frobenius automorphism for each exponent.
    static std::vector<std::vector<INDEX> > m_basis_representation; // Express the element as vector over the prime field.
    static unsigned int m_spacing; ///< The number of digits to represent an element.

    static void setSpacing_bitsPerEntry(void);
public:
    typedef SymWeight<1> MyWeight;

    static std::string myName(void)
    {
        std::stringstream sstr("GF(");
        sstr << m_q << ")";
        return sstr.str();
    }

    static unsigned int getSpacing(void)
    {
        return m_spacing;
    }
    static unsigned int chainlength(void)
    {
        return 1;
    }
    static unsigned int bitsPerEntry(void)
    {
        return m_bitsPerEntry;
    }

    static void init(unsigned int p);
    static void init(const char* const filename);
    static std::ostream& print(std::ostream& os);

    static My_Long_Int SubmoduleQty(const unsigned int* const type, unsigned int n)
    {
        return help::GaussBinomial(n, type[0], m_q);
    }

    inline static Element<Field> thetapow(unsigned int i);
    inline static unsigned int my_groundfield_size(void);
    inline static unsigned int my_size(void);
    inline static unsigned int my_multiplicativeGroupSize(void);
    inline static unsigned int get_r(void);
    inline static unsigned int get_primefield_size(void);
    inline static Element<Field> get_primitiveElement(void);
    inline static Element<Field> random(void);
    inline static Element<Field> randomInvertible(void);
    inline static unsigned int height(const Element<Field> a);
    inline static std::vector<Element<Field> > multGenerator(void);
    inline static std::vector<Element<Field> > nontrivalUnits(void);
    static std::vector<Element<Field> > nontrivalUnitsPlusMinus(void);

    inline static Element<Field> add(const Element<Field> a, const Element<Field> b);
    inline static Element<Field> addInverse(const Element<Field> a);
    inline static Element<Field> minus(const Element<Field> a, const Element<Field> b);
    inline static std::ostream& print(std::ostream& os, const Element<Field> a);

    inline static bool isUnit(const Element<Field> a);
    inline static Element<Field> mult(const Element<Field> a, const Element<Field> b);
    inline static Element<Field> multInverse(const Element<Field> a); // returns 0 if nilpotent
    inline static Element<Field> div(const Element<Field> a, const Element<Field> b);
    inline static int compare(const Element<Field> a, const Element<Field> b);

    inline static const std::vector<unsigned int>& getFrobenius(unsigned int frob_pow);
    inline static const std::vector<unsigned int>& getBasisRepresentation(const Element<Field> a);

    inline static Element<Field>primeFieldElement(unsigned int);

    // things we should not uses
    //	inline static bool isReducable(const Element<Field> a);
};

class BinaryField
{
public:
    typedef SymWeight<1> MyWeight;
    static std::string myName(void)
    {
        return "GF(2)";
    }

    static unsigned int getSpacing(void)
    {
        return 2;
    }
    static unsigned int chainlength(void)
    {
        return 1;
    }
    static unsigned int bitsPerEntry(void)
    {
        return 1;
    }

    static My_Long_Int SubmoduleQty(const unsigned int* const type, unsigned int n)
    {
        return help::GaussBinomialBinary(n, type[0]);
    }

    inline static Element<BinaryField> thetapow(unsigned int i)
    {
        return 1;
    }
    inline static unsigned int my_groundfield_size(void)
    {
        return 2;
    }
    inline static unsigned int my_size(void)
    {
        return 2;
    }
    inline static unsigned int my_multiplicativeGroupSize(void)
    {
        return 1;
    }
    inline static unsigned int get_r(void)
    {
        return 1;
    }
    inline static unsigned int get_primefield_size(void)
    {
        return 2;
    }
    inline static Element<BinaryField> get_primitiveElement(void)
    {
        return 1;
    }
    inline static Element<BinaryField> random(void)
    {
        return rand() % 2;
    }
    inline static Element<BinaryField> randomInvertible(void)
    {
        return 1;
    }
    inline static unsigned int height(const Element<BinaryField> a)
    {
        return 1 - ((int) a.getIndex());
    }
    //   inline static std::vector< Element<BinaryField> > multGenerator(void);

    inline static Element<BinaryField> add(const Element<BinaryField> a, const Element<BinaryField> b)
    {
        return a.getIndex() ^ b.getIndex();
    }
    inline static Element<BinaryField> addInverse(const Element<BinaryField> a)
    {
        return a;
    }
    inline static Element<BinaryField> minus(const Element<BinaryField> a, const Element<BinaryField> b)
    {
        return a.getIndex() ^ b.getIndex();
    }
    inline static std::ostream& print(std::ostream& os, const Element<BinaryField> a)
    {
        return os << a.getIndex();
    }

    inline static bool isUnit(const Element<BinaryField> a)
    {
        return a.getIndex() == 1;
    }
    inline static Element<BinaryField> mult(const Element<BinaryField> a, const Element<BinaryField> b)
    {
        return a.getIndex() & b.getIndex();
    }
    inline static Element<BinaryField> multInverse(const Element<BinaryField> a)
    {
        return a;
    }
    inline static Element<BinaryField> div(const Element<BinaryField> a, const Element<BinaryField> b)
    {
        return a.getIndex() & b.getIndex();
    }
    inline static int compare(const Element<BinaryField> a, const Element<BinaryField> b)
    {
        return (int) a.getIndex() - (int) b.getIndex();
    }
};


Element<Field> Field::thetapow(unsigned int i)
{
    if (i == 0)
    {
        return 1;
    }
    std::cerr << "An error occured in  Z<p>::thetapow (index is bigger than allowed)" << std::endl;
    exit(-1);
}

unsigned int Field::my_groundfield_size(void)
{
    return m_q;
}

unsigned int Field::my_size(void)
{
    return m_q;
}

unsigned int Field::my_multiplicativeGroupSize(void)
{
    return m_q - 1;
}

unsigned int Field::get_r(void)
{
    return m_r;
}

unsigned int Field::get_primefield_size(void)
{
    return m_p;
}

Element<Field> Field::get_primitiveElement(void)
{
    return m_alpha;
}

Element<Field> Field::random(void)
{
    return rand() % m_q;
}

Element<Field> Field::randomInvertible(void)
{
    return rand() % (m_q - 1) + 1;
}

std::vector<Element<Field> > Field::multGenerator(void)
{
    if (m_q == 2)
    {
        return std::vector<Element<Field> >(0); // the multiplicative group is generated by the empty set...
    }
    return std::vector<Element<Field> >(1, m_alpha);
}

std::vector<Element<Field> > Field::nontrivalUnits(void)
{
    std::vector<Element<Field> > res(m_q - 2, 0);
    std::vector<Element<Field> >::iterator resIter(res.begin());
    std::vector<Element<Field> >::iterator resIterEnd(res.end());
    INDEX i = 2;
    for (; resIter < resIterEnd; ++resIter, ++i)
    {
        *resIter = i;
    }
    return res;
}

Element<Field> Field::add(const Element<Field> a, const Element<Field> b)
{
    return m_addition[a.getIndex() * m_q + b.getIndex()];
}

Element<Field> Field::addInverse(const Element<Field> a)
{
    return m_additive_inv[a.getIndex()];
}

Element<Field> Field::minus(const Element<Field> a, const Element<Field> b)
{
    return m_subtraction[a.getIndex() * m_q + b.getIndex()];
}

std::ostream& Field::print(std::ostream& os, const Element<Field> a)
{
    return os << ((unsigned int) a.getIndex());
}

bool Field::isUnit(const Element<Field> a)
{
    return a.getIndex() != 0;
}

Element<Field> Field::mult(const Element<Field> a, const Element<Field> b)
{
    return m_multiplication[a.getIndex() * m_q + b.getIndex()];
}

Element<Field> Field::multInverse(const Element<Field> a)
{
    return m_multiplicative_inv[a.getIndex()];
}

Element<Field> Field::div(const Element<Field> a, const Element<Field> b)
{
    return m_division[a.getIndex() * m_q + b.getIndex()];
}

// returns -1 if a<b, +1 if a>b and 0 if a=b
int Field::compare(const Element<Field> a, const Element<Field> b)
{
    if (a.getIndex() == b.getIndex())
    {
        return 0;
    }
    if (a.getIndex() < b.getIndex())
    {
        return -1;
    }
    return 1;
}

unsigned int Field::height(const Element<Field> a)
{
    return (a.getIndex() == 0);
}

const std::vector<unsigned int>& Field::getFrobenius(unsigned int frob_pow)
{
    return m_frobenius[frob_pow];
}

const std::vector<unsigned int>& Field::getBasisRepresentation(const Element<Field> a)
        {
    return m_basis_representation[a.getIndex()];
        }

Element<Field> Field::primeFieldElement(unsigned int i)
{
    return Element<Field>(m_primeFieldElements[i]);
}

// things we should not uses
//bool Field::isReducable(const Element<Field> a)
//{
//	std::cerr<<"We should not need Field::isReducable() for finite fields"<<std::endl;
//	exit(-1);
//	return a.getIndex()>1;
//}

#endif /* FIELD_HPP_ */
