/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  gennode.hpp
///  \brief Implements the backtracking procedure.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef GENNODE_H_
#define GENNODE_H_

#include "node.hpp"
#include "gen.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class GenNode
/// \brief The backtracking through the symmetric group.
///
/// This implements the outer canonization algorithm described in
/// <em>T. Feulner, Canonization of linear codes over \f$\mathbb{Z}_4\f$, AMC, submitted</em> and
/// <em> T. Feulner; The automorphism groups of  linear codes and canonical representatives of their semilinear isometry classes, AMC 3(4), pp. 363 - 383
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename RING, AlgTypeE ALGTYPE>
class GenNode: public generator<Node<RING, ALGTYPE> >
{
private:
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /// \class PredecessorState
    /// \brief The necessary things we should know about the predecessors of the actual node.
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    class PredecessorState
    {
    public:
        unsigned int m_brothers; ///< How many brothers there are for the predecessor node.
        unsigned int m_index; ///< The son indeces of the predecessor.
        Permutation m_perm_from_root; ///< The product of the transversal elements.
        OrdPartition m_block_signature; ///< The blocks of the predecessor node.

        /// One refinement method works on an invariant subset of codewords, this gives the size in one of the predecessor levels.
        unsigned int m_minWord_workSize;
        std::vector<bool> m_eliminated_pos; ///< The invariant subset is reduced by shortening on some positions. This gives these positions
        std::vector<unsigned int> m_fixedUnEliminated; ///< The positions which are fixed, but the information is not deleted by shortening.
        bool m_isInSpan; ///< The type of minimization.
        InnerStab<RING, ALGTYPE> m_innerStab; ///< The inner stabilizer.
    };

public:
    int m_last_lvl; ///< Save the level of the last node.
    unsigned int m_cols; ///< Parameter n of the code.
    unsigned int m_k; ///< Parameter k of the code.
    unsigned int m_k0; ///< Parameter k0 of the code.

    std::vector<PredecessorState> m_predStates; ///< All important data for the backtracking.
    std::vector<RefinementValues> m_bestRefinement; ///< The results of the refinement for the actual candidate for the canonical representative.
    Matrix<RING> m_GammaBest; ///< The actual candidate for the canonical representative.
    bool m_isBestInitialized; ///< Is there are candidate, or are we going to replace it with the next leaf node.
    std::vector<int> m_autom_labra_father; ///< The father vector of the Labelled Branching for the known automorphisms.

    bool m_only_automs;

    // The virtual function which has to be implemented.
    /// \brief Get depth of the backtrack tree, viz the number of columns.
    int depth() const
    {
        return m_cols;
    }

    // The virtual function which has to be implemented.
    /// Get first son of current node (on level \a lvl) which is allowed (i.e. is at least as good as the candidate).
    bool gen_first(int lvl);

    // The virtual function which has to be implemented
    /// Get the next node on level \a lvl which is allowed (i.e. is at least as good as the candidate).
    bool gen_next(int level);

    /// \brief Test if the algorithm can skip the subtree, without any loss of information.
    /// \sa T. Feulner, The automorphism groups of linear codes and canonical representatives of their semilinear isometry classes,
    /// AMC, 3, pp. 363 - 383, Nov 2009
    bool labraTest(const Permutation& perm) const;
public:
    /// Constructor, \a root defines the root node of the backtrack tree.
    GenNode(unsigned int k_0, const Node<RING, ALGTYPE>& root, bool only_automs) :
        generator<Node<RING, ALGTYPE> > (root), m_cols(root.getCols()), m_k(root.getRows()), m_k0(k_0), //m_root(root),
                m_predStates(m_cols), m_bestRefinement(m_cols), m_only_automs(only_automs)
    {
        this->m_isBestInitialized = false;
    }
    ~GenNode(void)
    {
    }

    /// Is there a candidate for the canonical representative at the moment.
    bool getBestInitialized(void) const
    {
        return this->m_isBestInitialized;
    }

    /// Set the candidate for the canonical representative to the actual node (should only be called for leaf nodes).
    void setBestInitialized()
    {
        this->m_isBestInitialized = true;
        m_GammaBest = this->m_sol.getMatrix();
    }

    /// The father vector of the labelled branching is replaced.
    void resetFather(const std::vector<int>& autom_labra_father)
    {
        m_autom_labra_father = autom_labra_father;
    }

    /// Get the permutation from the root to the actual node.
    const Permutation& getPermutation(void) const
    {
        return this->m_sol.getPermutation();
    }

    /// Compare the last fixed with the corresponding column of the candidate (if it exist).
    bool compareCols(unsigned int index);

    /// Check if the the projection onto the fixed columns has full row rank, and if true if the matrix has a discrete block partition.
    bool HasDiscreteBlocks(void);

    /// Minimize the unfixed columns of the actual node in the order they occur.
    bool doDiscreteCompletion(void);
};

template<typename RING, AlgTypeE ALGTYPE>
bool GenNode<RING, ALGTYPE>::compareCols(unsigned int index)
{
    if (this->m_isBestInitialized)
    {
        const Vec<RING>& actCol = this->m_sol.getMatrix().getColumn(index);
        const Vec<RING>& bestCol = m_GammaBest.getColumn(index);

        if (m_only_automs)
        {
            if (actCol != bestCol)
            {
                return false;
            }
        }
        else
        {
            int cmp(actCol.compare(bestCol));
            if (cmp<0)
            {
                this->m_isBestInitialized = false;
                return true;
            }
            if (cmp>0)
            {
                return false;
            }
        }
    }
    return true;
}

template<typename RING, AlgTypeE ALGTYPE>
bool GenNode<RING, ALGTYPE>::gen_first(int lvl)
{
    if (lvl >= (int) m_cols)
    {
        return false;
    }
    m_last_lvl = lvl;
    NextFixInfo refRes(0, m_cols, false);

    if (lvl > 0)
    {
        // which column should be fixed next and how big is this block
        refRes = this->m_sol.calcNextFixInfo(m_bestRefinement[lvl - 1].m_heights);
    }
    else
    {
        // refine the root node and then decide like above. We do not need to store the result, since there is only one node on level 0.
        RefinementValues dummy;
        bool xyz(false);
        this->m_sol.refine(lvl - 1, dummy, xyz, m_k0, this);
        refRes = this->m_sol.calcNextFixInfo(dummy.m_heights);
    }

    this->m_sol.fixCol(refRes.m_index);
    PredecessorState& storeStates = m_predStates[lvl];
    storeStates.m_innerStab = this->m_sol.getInnerStab();
    storeStates.m_index = 0;
    storeStates.m_brothers = refRes.m_blockSize;
    storeStates.m_isInSpan = refRes.m_isInSpan;
    storeStates.m_block_signature = this->m_sol.getBlockSignature();
    storeStates.m_perm_from_root = this->m_sol.getPermutation();
    storeStates.m_minWord_workSize = this->m_sol.getMinWord_workSize();
    storeStates.m_eliminated_pos = this->m_sol.getEliminatedPos();
    storeStates.m_fixedUnEliminated = this->m_sol.getfixedUnEliminated();

    if (labraTest(storeStates.m_perm_from_root))
    {
        // this subtree does not contain any new information
        this->m_latex.prunedNode("labra");
        return gen_next(lvl);
    }

    this->m_sol.minimize(m_k0, refRes.m_isInSpan); // minimize the fixed column with the inner stabilizer

    if (!compareCols(refRes.m_index))
    {
        // the best candidate has a smaller column
        this->m_latex.prunedNode("lexOrder");
        return gen_next(lvl);
    }

    if (!this->m_sol.refine(lvl, m_bestRefinement[lvl], this->m_isBestInitialized, m_k0, this))
    {
        // the best candidate has better images under the homomorphisms
        return gen_next(lvl);
    }

    return true;
}

template<typename RING, AlgTypeE ALGTYPE>
bool GenNode<RING, ALGTYPE>::gen_next(int lvl)
{
    if (lvl >= (int) m_cols)
    {
        return false;
    }

    PredecessorState& pre_state = m_predStates[lvl];
    unsigned int colToFix = this->m_sol.getFixedCol(lvl);

    while (true)
    {
        /// restore the state on lvl
        this->m_sol.backTo(lvl, pre_state.m_minWord_workSize, pre_state.m_eliminated_pos,
                pre_state.m_fixedUnEliminated, pre_state.m_block_signature, pre_state.m_innerStab);
        ++pre_state.m_index; // try the next brother
        if (pre_state.m_brothers <= pre_state. m_index)
        {
            // we have visited all brothers
            return false;
        }

        Permutation::mulTransposition(colToFix, colToFix + pre_state.m_index, pre_state.m_perm_from_root);
        Permutation perm;
        Permutation::inv(this->m_sol.getPermutation(), perm);
        Permutation::mul(pre_state.m_perm_from_root, perm, perm);
        this->m_sol.permuteMatrix(perm, &pre_state.m_perm_from_root); // restore the ordering of the columns
        m_last_lvl = lvl;

        if (labraTest(pre_state.m_perm_from_root))
        {
            // this subtree does not contain any new information
            this->m_latex.prunedNode("labra");
            continue; // try the next brother
        }

        this->m_sol.minimize(m_k0, pre_state.m_isInSpan);

        if (!compareCols(colToFix))
        {
            // the best candidate has a smaller column
            this->m_latex.prunedNode("lexOrder");// this node did not lead to a canonical element, print it to the latexFile
            continue; // try the next brother
        }

        if (this->m_sol.refine(lvl, m_bestRefinement[lvl], this->m_isBestInitialized, m_k0, this))
        {
            // this node is as least as good as the candidate
            return true;
        }

        // this node did not lead to a canonical element, print it to the latexFile
        this->m_latex.prunedNode("refine");
    }
    return false;
}

template<typename RING, AlgTypeE ALGTYPE>
bool GenNode<RING, ALGTYPE>::labraTest(const Permutation& perm) const
{
    if (m_autom_labra_father.size() == 0)
    {
        return false;// trivial automorphism group
    }

    unsigned int k;

    std::vector<int>::const_iterator blocks_begin = this->m_sol.getBlockSignature().begin(), blocks_it;
    std::vector<bool>::const_iterator m_isFixedCol_it = this->m_sol.get_isFixedCol().begin() + 1;
    std::vector<int>::const_iterator father_it = m_autom_labra_father.begin();
    int deg = perm.degree();

    //	if(m_can->getOptions().m_labraTightened)
    //	{
    unsigned int m1, m2;
    for (k = 0; k < m_cols; ++k)
    {
        if (*father_it > -1)
        {
            m1 = perm.apply(k, deg);
            m2 = perm.apply(*father_it, deg);

            if (m2 < m1)
            {
                ++father_it;
                continue; // lexicographically ordered
            }

            // m1 < m2
            ++m1;
            blocks_it = blocks_begin + m1;
            while (m1 < m2 && !*(blocks_it))
            {
                ++m1;
                ++blocks_it;
            }
            if (m1 < m2 || *blocks_it)
            {
                return true;
            }
        }
        ++father_it;
    }
    return false;
}

template<typename RING, AlgTypeE ALGTYPE>
bool GenNode<RING, ALGTYPE>::HasDiscreteBlocks(void)
{
    //	const InnerStab<RING, ALGTYPE>& actStab = this->m_sol.getInnerStab();
    //	if(actStab.m_rank < m_k0)
    //	{
    //		return false;
    //	}

    const OrdPartition& blocks = this->m_sol.getBlockSignature();
    std::vector<int>::const_iterator blocksIter = blocks.begin();
    std::vector<int>::const_iterator blocksIterEnd = blocks.end();

    for (; blocksIter < blocksIterEnd; ++blocksIter)
    {
        if (*blocksIter == 0)
        {
            return false;
        }
    }

    return true;
}

template<typename RING, AlgTypeE ALGTYPE>
bool GenNode<RING, ALGTYPE>::doDiscreteCompletion(void)
{
    const InnerStab<RING, ALGTYPE>& innerStab(this->m_sol.getInnerStab());
    unsigned int missingForK0(0);
    bool searchK0(false);
    if (RING::chainlength() == 1 || m_k0 <= innerStab.m_rank)
    {
        missingForK0 = 0;
    }
    else
    {
        missingForK0 = m_k0 - innerStab.m_rank;
        searchK0 = true;
    }

    const std::vector<bool>& fixedCols = this->m_sol.get_isFixedCol();
    for (unsigned int col_index = 0; col_index < m_cols;)
    {
        if (!fixedCols[col_index])
        {
            unsigned int myHeight = this->m_sol.getMatrix().getColumn(col_index).height(innerStab.m_rank).first;
            if (!searchK0 || myHeight != 1)
            {
                this->m_sol.fixCol(col_index);
                this->m_sol.minimize(m_k0, myHeight == RING::chainlength());
                if (!compareCols(col_index))
                {
                    this->m_latex.prunedNode("lexOrder");
                    return false;
                }
                if (searchK0 && myHeight == 0)
                {
                    missingForK0--;
                    if (missingForK0 == 0)
                    {
                        col_index = 0;
                        searchK0 = false;
                        continue; // don't perform ++col_index in the for loop!
                    }
                }
            }
        }
        ++col_index;
    }
    return true;
}

#endif /*GENNODE_H_*/
