/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  main.cpp
///  \brief Starting routines.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   23.11.2012
///
///  Copyright (C) 2012 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <vector>

#include "f2_f2.hpp"
#include "z4.hpp"
#include "field.hpp"
#include "codecan.hpp"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <dirent.h>

template<typename RING, AlgTypeE ALGTYPE>
void printResult(const CodeCan<RING, ALGTYPE>& canonizer, const Matrix<RING>& mat, const OrdPartition* const partitionInit=NULL)
{
    std::cout << "The canonical representative is" << std::endl;
    if(partitionInit != NULL)
    {
        std::cout <<Node<RING, ALGTYPE>(mat.getRows(), canonizer.getGammaBest(), partitionInit)<<std::endl;
    }
    else
    {
        std::cout <<canonizer.getGammaBest()<<std::endl;
    }
    std::cout << "The transporter" << std::endl;
    std::cout << canonizer.getTransporter() << std::endl;

    if ( canonizer.getTransporter() * mat != canonizer.getGammaBest() )
    {
        std::cout<<"The transporter element is wrong! Please contact Thomas Feulner."<<std::endl;
        std::cout<<canonizer.getTransporter() <<std::endl<< (canonizer.getTransporter() * mat) <<std::endl;;
    }

    Matrix<RING> gammaBest = canonizer.getGammaBest();
    std::cout << "The generators of the automorphism group are" << std::endl;
    const std::vector<ActingElement<RING, ALGTYPE> >& automs = canonizer.getAutoms();
    for (unsigned int i = 0; i < automs.size(); ++i)
    {
        std::cout << automs[i] << std::endl;
        if( (automs[i] * mat) != mat )
        {
            std::cout<<"is wrong Automorphism!!!!!!!!!!!!!!!!! Please contact Thomas Feulner."<<std::endl;
            std::cout<<automs[i] * mat << std::endl;
        }
    }

    std::pair<My_Long_Int, My_Long_Int> autOrder(canonizer.getAutomOrder());
    std::cout << "The order of the automorphism group is " << autOrder.first <<" "<< autOrder.second <<" (\"nontrivial\") "<< std::endl;
}

template<typename RING>
void readFile(const std::string& filename, Matrix<RING>& mat, OrdPartition& blocks, std::set<unsigned int>& blockstarts)
{
    std::ifstream in;
    in.open(filename.c_str());
    if (!in)
    {
        std::cerr << "Unable to open input file." << std::endl;
        exit(1);
    }

    unsigned int n(0), k(0);
    std::vector<unsigned int> blocksizes;
    unsigned int dummy;

    in>>std::ws;
    while((unsigned char) in.peek() == '#')
    {
        in.ignore (std::numeric_limits<std::streamsize>::max(), '\n' );
    }
    std::string line;
    getline( in, line );
    std::stringstream linestream(line);

    do
    {
        linestream>>dummy;
        linestream>>std::ws;
        n+=dummy;
        blocksizes.push_back(dummy);
    }while(linestream.good());

    blocks.clear();
    blocks.resize(n);
    std::vector<unsigned int>::const_iterator blocksizesIter(blocksizes.begin());
    blockstarts.clear();
    blockstarts.insert(0);

    unsigned int count(0);
    while(count<n)
    {
        blocks[count] =1;
        count += *blocksizesIter;
        blockstarts.insert(count);
        ++blocksizesIter;
    }

    in>>std::ws;
    while((unsigned char) in.peek() == '#')
    {
        in.ignore (std::numeric_limits<std::streamsize>::max(), '\n' );
    }
    in>>k;
    in>>std::ws;
    mat.reset(k,n,in);
    in.close();
}


template<typename RING, AlgTypeE ALGTYPE>
void canonize(CodeCan<RING, ALGTYPE>& canonizer, const std::string& filename)
{
    Matrix<RING> mat(filename.c_str());
    std::cout << "We are canonizing the following matrix:" << std::endl;
    std::cout << mat << std::endl;
    CodeCanStarter<ALGTYPE>::start(canonizer, mat, false);
    printResult(canonizer, mat);
}

template<typename RING, AlgTypeE ALGTYPE>
void allinDir(CodeCan<RING, ALGTYPE>& canonizer, const std::string& prefix)
{
    std::map<Matrix<RING>, std::vector<std::string> > canonicals;
    DIR *hdir;
    struct dirent *entry;
    unsigned int count(0);

    hdir = opendir(prefix.c_str());

    do
    {
        entry = readdir(hdir);
        if (entry && entry->d_type == DT_REG)
        {
            ++count;

            Matrix<RING> mat;
            OrdPartition blocks;
            std::set<unsigned int> blockstarts;
            readFile( prefix + entry->d_name , mat, blocks, blockstarts);
            std::vector<unsigned int> shape( mat.toOrderedBasis(false, NULL, NULL) );

            unsigned int truedim = shape[0]+ shape[1];
            for(; truedim<mat.getRows(); )
            {
                mat.removeLastRow();
            }

            Node<RING, ALGTYPE> root(shape[0], mat, &blocks);
            CodeCanStarter<ALGTYPE>::advanced_start(shape[0], canonizer, root, false,  std::vector<Permutation>());

            std::vector<std::string>& fnames(canonicals[canonizer.getGammaBest()]);
            if(fnames.empty())
            {
                std::cout<<"New Representative: "<<std::endl;
                std::cout<<entry->d_name <<std::endl;
                std::cout<<canonizer.getGammaBest()<<std::endl;
                std::cout<<canonizer.getAutomOrder().first<< " "<<canonizer.getAutomOrder().second<< " (\"nontrivial\")" <<std::endl;

                const std::vector<ActingElement<RING, ALGTYPE> > autGens(canonizer.getAutoms() );
                help::print(std::cout, autGens, "\n")<<std::endl;
            }
            fnames.push_back( entry->d_name );
        }
    } while (entry);
    closedir(hdir);

    std::cout<<canonicals.size()<<" isomorphism classes from input size: "<<count<<std::endl;
    typename std::map<Matrix<RING>, std::vector<std::string> >::const_iterator mapIter(canonicals.begin());
    typename std::map<Matrix<RING>, std::vector<std::string> >::const_iterator mapIterEnd(canonicals.end());

    std::cout<<"Filenames sorted by isomorphism classes"<<std::endl;
    for( ; mapIter!=mapIterEnd; ++mapIter)
    {
        help::print(std::cout, mapIter->second, " ")<<std::endl<<std::endl;
    }

    return;
}

template<typename RING, AlgTypeE ALGTYPE>
void advanced(CodeCan<RING, ALGTYPE>& canonizer, const std::string& filename, const std::string* const autom_file)
{
    Matrix<RING> mat;
    OrdPartition blocks;
    std::set<unsigned int> blockstarts;
    readFile<RING>(filename, mat, blocks, blockstarts);

    Matrix<RING> matCopy(mat);
    std::vector<unsigned int> blocksize(matCopy.toOrderedBasis(false, NULL, NULL));
    matCopy.normalizeColumns(ALGTYPE);
    unsigned int k0(blocksize[0]);

    unsigned int rank(k0);
    std::vector<unsigned int>::const_iterator blocksizeIter(blocksize.begin());
    std::vector<unsigned int>::const_iterator blocksizeIterEnd(blocksize.end());
    ++blocksizeIter;
    for(; blocksizeIter<blocksizeIterEnd; ++blocksizeIter)
    {
        rank += *blocksizeIter;
    }

    if(rank != mat.getRows())
    {
        std::cout<<"You have to pass a matrix with full row rank!"<<std::endl;
        return;
    }

    Node<RING, ALGTYPE> root(k0, mat, &blocks);
    std::cout << "We are canonizing the following matrix:" << std::endl;
    std::cout << root << std::endl;

    std::vector<Permutation> known_automs;
    if(autom_file !=0)
    {
        std::ifstream in;
        std::cout<<"This is the list of automorphisms passed to the program:"<<std::endl;
        in.open(autom_file->c_str());
        if (!in)
        {
            std::cerr << "Unable to open automorphism input file." << std::endl;
            exit(1);
        }

        unsigned char c;
        in>>c;

        if(c == 'p')
        {
            unsigned int offset;
            in>>offset;
            Permutation dummy;
            while( dummy.scan(in, offset) )
            {
                known_automs.push_back(dummy);
                std::cout<<dummy<<std::endl;
            }
        }
        else if (c=='m')
        {
            Matrix<RING> A;
            Permutation pi_mat, pi_matPermuted;
            std::vector<Vec<RING> > cols( mat.getEntries());
            std::vector<Vec<RING> > colspermuted;

            typename std::vector<Vec<RING> >::iterator colsIter(cols.begin());
            typename std::vector<Vec<RING> >::iterator colsIterEnd(cols.end());
            if(ALGTYPE != PERMUTATIONAL)
            {
                for( ; colsIter<colsIterEnd; ++colsIter)
                {
                    colsIter->normalize(ALGTYPE==PLUSMINUSMONOMIAL);
                }
            }
            pi_mat.stable_sorting_perm(cols.begin(), colsIterEnd);
            pi_mat.permute(cols.begin(), colsIterEnd);
            Matrix<RING> matCopy(cols);

            Permutation::inv(pi_mat, pi_mat);
            in>>std::ws;
            while(in.good())
            {
                A.reset(rank,rank,in);
                Matrix<RING> matPermuted(A*mat);
                colspermuted = matPermuted.getEntries();
                colsIterEnd= colspermuted.end();

                if(ALGTYPE != PERMUTATIONAL)
                {
                    colsIter = colspermuted.begin();
                    for( ; colsIter<colsIterEnd; ++colsIter)
                    {
                        colsIter->normalize(ALGTYPE==PLUSMINUSMONOMIAL);
                    }
                }
                pi_matPermuted.stable_sorting_perm(colspermuted.begin(), colsIterEnd);
                pi_matPermuted.permute(colspermuted.begin(), colsIterEnd);
                matPermuted = Matrix<RING>(colspermuted);

                if( matPermuted != matCopy)
                {
                    std::cout<<"This matrix does not define an automorphism of the code:"<<std::endl;
                    std::cout<<A;
                }
                else
                {
                    Permutation aut( pi_mat*pi_matPermuted);
                    std::cout<<A;
                    std::cout<<"resulting permutation on the columns: "<< aut <<std::endl<<std::endl;
                    known_automs.push_back( aut );
                }
                in>>std::ws;
            }
        }
        else
        {
            std::cout<<"The file does not contain any description how the automorphisms are represented"<<std::endl;
        }
        in.close();
    }

    std::vector<Permutation>::iterator known_automsIter(known_automs.begin());
    while( known_automsIter != known_automs.end())
    {
        bool is_valid(true);
        unsigned int deg(known_automsIter->degree()), img;
        std::set<unsigned int>::const_iterator lb(blockstarts.begin());
        std::set<unsigned int>::const_iterator ub(lb);
        ++ub;

        // check if this permutation respects the partitioning of the columns
        for(unsigned int i=0; i<mat.getCols(); ++i)
        {
            if( i == *ub)
            {
                ++lb;
                ++ub;
            }
            img = known_automsIter->apply((int) i, (int) deg);
            if(img<*lb || img>=*ub)
            {
                std::cout<<"The permutation"<< *known_automsIter <<"permutes entries from different blocks"<<std::endl;
                is_valid = false;
                break;
            }
        }

        if(is_valid)
        {
            ++known_automsIter;
        }
        else
        {
            known_automsIter = known_automs.erase(known_automsIter);
        }
    }

    CodeCanStarter<ALGTYPE>::advanced_start(k0, canonizer, root, false,  known_automs);
    printResult(canonizer, mat, &root.getBlockSignature());
}

template<typename RING, AlgTypeE ALGTYPE>
int call_main(int argc, char** argv, unsigned int argument_startPos)
{
#ifdef USE_MAGMA
    char magma_call[] =MAGMA_CALL;
    char* magma_cas_call[2];
    magma_cas_call[0] = magma_call;
    magma_cas_call[1] = 0;
    MagmaCASModule* const cas = new MagmaCASModule(magma_cas_call, CASPATH);
#else
    char gap_call[] = GAP_CALL;
    char silence_option[] = "-q";
    char* gap_cas_call[3];
    gap_cas_call[0] = gap_call;
    gap_cas_call[1] = silence_option;
    gap_cas_call[2] = 0;
    GapCASModule* const cas = new GapCASModule(gap_cas_call, CASPATH) ;
#endif


    CodeCan<RING, ALGTYPE> canonizer((AbstractCASModule* const ) cas);
    const std::string method(argv[argument_startPos]);

    if (method == "-allinDir")
    {
        allinDir<RING, ALGTYPE> (canonizer, argv[argument_startPos + 1]);
    }
    else if(method == "-canonize")
    {
        canonize<RING, ALGTYPE>(canonizer,  argv[argument_startPos + 1]);
    }
    else if(method == "-advanced")
    {
        if((unsigned int) argc>argument_startPos + 2)
        {
            std::string automfile(argv[argument_startPos + 2]);
            advanced<RING, ALGTYPE> (canonizer, argv[argument_startPos + 1], &automfile);
        }
        else
        {
            advanced<RING, ALGTYPE> (canonizer, argv[argument_startPos + 1], NULL);
        }
    }
    else
    {
        std::cout<<"You passed no valid method argument"<<std::endl;
    }

    delete cas;

    return 0;
}

template<typename RING>
int AlgTypeDecision(int argc, char** argv, unsigned int argument_startPos)
{
    if (std::string(argv[1]) == "permutational")
    {
        return call_main<RING, PERMUTATIONAL>(argc, argv,  argument_startPos);
    }
    if (std::string(argv[1]) == "plusminusmonomial")
    {
        std::cout<<"not implemented for arbitrary rings"<<std::endl;
        return -1;
    }
    return call_main<RING, LINEAR>(argc, argv,  argument_startPos);
}



template<>
int AlgTypeDecision<BinaryField>(int argc, char** argv, unsigned int argument_startPos)
{
    return call_main<BinaryField, PERMUTATIONAL>(argc, argv,  argument_startPos);
}

template<>
int AlgTypeDecision<Field>(int argc, char** argv, unsigned int argument_startPos)
{
    if (std::string(argv[1]) == "permutational")
    {
        return call_main<Field, PERMUTATIONAL>(argc, argv,  argument_startPos);
    }
    if (std::string(argv[1]) == "plusminusmonomial")
    {
        if( Field::get_primefield_size() == 2)
        {
            std::cerr<<"The characteristic of the field is two. Please use option \"permutational\"."<<std::endl;
            return 0;
        }
        return call_main<Field, PLUSMINUSMONOMIAL>(argc, argv,  argument_startPos);
    }
    if (std::string(argv[1]) == "linear")
    {
        return call_main<Field, LINEAR>(argc, argv,  argument_startPos);
    }
    return call_main<Field, SEMILINEAR>(argc, argv,  argument_startPos);
}

int RingDecision(int argc, char** argv)
{
    if (std::string(argv[2]) == "Z4")
    {
        return AlgTypeDecision<Z4> (argc, argv, 3);
    }
    else if (std::string(argv[2]) == "Z2")
    {
        return AlgTypeDecision<BinaryField> (argc, argv, 3);
    }
    else if (std::string(argv[2]) == "F2_F2")
    {
        return AlgTypeDecision<F2_F2> (argc, argv, 3);
    }
    else if (argv[2][0] == 'Z')
    {
        Field::init(atoi(argv[2] + 1));
        return AlgTypeDecision<Field> (argc, argv, 3);
    }
    else  if (argv[2][0] == 'F')
    {
        Field::init( argv[3] );
        return AlgTypeDecision<Field> (argc, argv, 4);
    }
    else
    {
        std::cout << "No ring found" << std::endl;
        return 0;
    }
}


int main(int argc, char** argv)
{
    srand(time(NULL));
    DenseVec<BinaryField>::compute_bits_in_16bits();
    return RingDecision(argc, argv);
}

