/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  matrix.hpp
///  \brief Implements matrices over the ring R.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef MATRIX_HPP_
#define MATRIX_HPP_

#include <iomanip>
#include <limits>
#include <set>
#include "vector.hpp"
#include "z4.hpp"
#include "field.hpp"
#include "hashGenerator.h"
#include "partition.h"
#include "innerStab.hpp"
#include "magmaCASModule.h"
#include "gapCASModule.h"
//#include "dual.h"

template<typename RING>
class Matrix
{
public:
    Matrix(unsigned int k, unsigned int n, const DenseVec<BinaryField>& m); // complement to toIntRepresentation()
    Matrix(const char* const filename);
    Matrix(unsigned int k, unsigned int n, std::istream& in);
    Matrix(const std::vector<Vec<RING> >& mat);
    Matrix(const std::vector<std::vector<Element<RING> > >& mat);
    Matrix(unsigned int k = 0, unsigned int n = 0); // k \times n - matrix filled with zeros

    void reset(unsigned int k, unsigned int n, std::istream& in);

    void rowSubmatrix(const Matrix<RING>& G1, unsigned int begin, unsigned int nr_rows);
    void columnSubmatrix(unsigned int new_n);

    const Element<RING>& getElement(unsigned int i, unsigned int j) const;
    void setElement(unsigned int i, unsigned int j, const Element<RING>& el);
    unsigned int getRows(void) const;
    unsigned int getCols(void) const;
    const Vec<RING>& getColumn(unsigned int index) const
                                            {
        return m_entries[index];
                                            }
    const std::vector<Vec<RING> >& getEntries(void) const
                                            {
        return m_entries;
                                            }

    void swapRows(unsigned int row1, unsigned int row2);
    void swapCols(unsigned int col1, unsigned int col2);
    inline void permuteCols(const Permutation& perm);

    std::vector<unsigned int> heightSequence(unsigned int start_row = 0) const;

    void minimizeByColMult(bool only_minusOne, unsigned int row, UnordPartition& row_partition, const std::vector<unsigned int>& fixedCols,
            Matrix<RING>* transporterMat, Vec<RING>* transporterMult);
    void gaussElStep(unsigned int pivotRow, unsigned int pivotCol, bool doUpperEl, Matrix* const A = NULL);
    Matrix inverse(void) const;
    Matrix operator*(const Matrix& B) const;
    Vec<RING> operator*(const Vec<RING>& b) const;

    std::vector<unsigned int> toOrderedBasis(bool doUpperEl, Permutation* const trans, Matrix<RING>* const A); // returns shape
    std::vector<unsigned int> ColumnSpan(void);

    Matrix transpose(void) const;

    /// \brief Calculate all codewords of the linear code generated by this matrix.
    ///
    /// \todo Use the gray map.
    //	std::vector<Codeword<RING> >& getCodewords(std::vector<Codeword<RING> >& res, bool remove) const;

    void addRows(unsigned int fix, unsigned int to);
    void multiplyColumn(unsigned int col, const Element<RING>& a);
    void normalizeColumns(AlgTypeE algtype);
    void multiplyRow(unsigned int row, const Element<RING>& a);
    void multiplyRows(const std::vector<unsigned int>& rows, const Element<RING>& a,
            const std::vector<unsigned int>& notTouchCol = std::vector<unsigned int>());

    void appendZeroRow(void);
    void appendColumn(const Vec<RING>& newCol);
    void appendColumns(const std::vector<Vec<RING> >& newCols);
    void replaceLastColumn(const Vec<RING>& newCol);
    void replaceLastRow(const Vec<RING>& new_row);
    void removeLastColumn(void);
    void removeLastRow(void);
    unsigned int fillLastColProjective(unsigned int rank);

    template<typename R>
    friend std::ostream& operator<<(std::ostream& os, const Matrix<R>& mat);

    //    template<typename R>
    //    friend std::istream& operator<<(std::istream& in, Matrix<R>& mat);

    std::ostream& print(std::ostream& os, const char* sep, const std::string* const functionName) const;

    void identity(unsigned int k);
    void systematicDiagonal(unsigned int k0, unsigned int k1);

    bool operator!=(const Matrix& B) const;
    bool operator==(const Matrix& B) const;
    bool operator<(const Matrix& B) const;

    void random(unsigned int chainlength, unsigned int* type, unsigned int n);

    static unsigned int checkMatrix(const Matrix<RING>& Gamma, Matrix<RING>& Delta);

    // Span(G1) contains Span(G2), we want to have a basis of W such that  Span(G1) =  Span(G2) \oplus W
    static unsigned int complement(const Matrix<RING>& G1, const Matrix<RING>& G2, Matrix<RING>& res);
    static void VerticalJoin(const Matrix<RING>& G1, const Matrix<RING>& G2, Matrix<RING>& res);

    // only for finite fields:
    void applyFrobenius(unsigned int frob_pow, const std::vector<bool>* const isfixedCol = NULL);

//     unsigned int getDualMinDist(const unsigned int* const type, const KrawtchoukTable& krawtchouk) const;
//     unsigned int getMinDist(const unsigned int* const type, const KrawtchoukTable& krawtchouk) const;
//     bool verifyMinDist(const unsigned int* const type, const KrawtchoukTable& krawtchouk, unsigned int d, unsigned int desired_dual_d, bool residual) const;
    unsigned int calcMinDist(unsigned int desired_d) const;

    unsigned int weightOfLastRow(void) const;
private:
    unsigned int m_n, m_k;
    std::vector<Vec<RING> > m_entries; // columns of rows

    // matrix must be the transpose of an ordered basis matrix of type (n0, n-n0)
    //	std::vector<Codeword<RING> >& getColumnSpan(unsigned int n_0, std::vector<Codeword<RING> >& res, bool remove) const;
};

template<typename RING>
void Matrix<RING>::reset(unsigned int k, unsigned int n, std::istream& in)
{
    m_n = n;
    m_k = k;
    m_entries.resize(m_n, Vec<RING> (m_k, 0));
    unsigned int dummy;
    for (unsigned int j = 0; j < m_k; ++j)
    {
        in>>std::ws;
        while((unsigned char) in.peek() == '#')
        {
            in.ignore (std::numeric_limits<std::streamsize>::max(), '\n' );
        }
        for (unsigned int i = 0; i < m_n; ++i)
        {
            in >> dummy;
            m_entries[i].set(j, Element<RING> (dummy));
        }
    }
}

template<typename RING>
Matrix<RING>::Matrix(unsigned int k, unsigned int n, std::istream& in)
{
    reset(k, n, in);
}

template<typename RING>
Matrix<RING>::Matrix(const char* const filename)
{
    std::ifstream in;
    in.open(filename);

    if (!in)
    {
        std::cerr << "Unable to open input file." << std::endl;
        exit(1);
    }
    m_k = 0;
    m_n = 0;

    in>>std::ws;
    while((unsigned char) in.peek() == '#')
    {
        in.ignore (std::numeric_limits<std::streamsize>::max(), '\n' );
    }
    in >> m_n;

    in>>std::ws;
    while((unsigned char) in.peek() == '#')
    {
        in.ignore (std::numeric_limits<std::streamsize>::max(), '\n' );
    }
    in >> m_k;

    if (m_k <= 0)
    {
        std::cerr << "Nr of rows must be greater then 0." << std::endl;
        exit(1);
    }

    if (m_n <= 0)
    {
        std::cerr << "Number of columns must be greater then 0." << std::endl;
        exit(1);
    }

    reset(m_k, m_n, in);
    in.close();
}

template<typename RING>
Matrix<RING>::Matrix(const std::vector<Vec<RING> >& mat) :
m_n(mat.size()), m_k(0), m_entries(mat)
{
    if (m_n > 0)
    {
        m_k = mat[0].size();
    }
}

template<typename RING>
Matrix<RING>::Matrix(unsigned int k, unsigned int n) :
m_n(n), m_k(k), m_entries(std::vector<Vec<RING> >(n, Vec<RING> (k)))
{
}

template<typename RING>
Matrix<RING>::Matrix(const std::vector<std::vector<Element<RING> > >& mat) :
m_n(mat.size()), m_k(0)
{
    if (m_n > 0)
    {
        m_k = mat[0].size();
    }
    m_entries.resize(m_n);
    copy(mat.begin(), mat.end(), m_entries.begin());
}

template<typename RING>
const Element<RING>& Matrix<RING>::getElement(unsigned int i, unsigned int j) const
{
    return m_entries[j].get(i);
}

template<typename RING>
void Matrix<RING>::setElement(unsigned int i, unsigned int j, const Element<RING>& el)
{
    m_entries[j].set(i, el);
}

template<typename RING>
unsigned int Matrix<RING>::getRows(void) const
{
    return m_k;
}

template<typename RING>
unsigned int Matrix<RING>::getCols(void) const
{
    return m_n;
}

template<typename RING>
std::ostream& operator<<(std::ostream& os, const Matrix<RING>& mat)
{
    return mat.print(os, NULL, NULL);
}

template<typename RING>
std::ostream& Matrix<RING>::print(std::ostream& os, const char* sep, const std::string* const functionName ) const
{
    if (getRows() == 0 || getCols() == 0)
    {
        os << "()" << std::endl;
        return os;
    }
    for (unsigned int i = 0; i < getRows(); ++i)
    {
        if(i != 0 && sep != 0)
        {
            os<<sep;
        }
        for (unsigned int j = 0; j < getCols(); ++j)
        {
            if(j != 0 && sep != 0)
            {
                os<<sep;
            }
            if(functionName != NULL)
            {
                os<<*functionName<<"("<<getElement(i, j) <<")";
            }
            else
            {
                os.width(RING::getSpacing());
                os.flags(std::ios::right);
                os.fill(' ');
                os << getElement(i, j);
            }
        }

        os << std::endl;
    }
    return os;
}

template<typename RING>
std::istream& operator>>(std::istream& in, Matrix<RING>& mat)
{
    INDEX dummy;
    for (unsigned int i = 0; i < mat.getRows(); ++i)
    {
        for (unsigned int j = 0; j < mat.getCols(); ++j)
        {
            in >> dummy;
            mat.setElement(i, j, dummy);
        }
    }
    return in;
}

template<typename RING>
bool Matrix<RING>::operator==(const Matrix<RING>& B) const
{
    typename std::vector<Vec<RING> >::const_iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(m_entries.end());
    typename std::vector<Vec<RING> >::const_iterator BcolIter(B.m_entries.begin());

    for (; colIter < colIterEnd; ++colIter, ++BcolIter)
    {
        if (*colIter != *BcolIter)
        {
            return false;
        }
    }

    return true;
}

template<typename RING>
bool Matrix<RING>::operator!=(const Matrix<RING>& B) const
{
    return !(*this == B);
}

template<typename RING>
bool Matrix<RING>::operator<(const Matrix<RING>& B) const
{
    typename std::vector<Vec<RING> >::const_iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::const_iterator BcolIter(B.m_entries.begin());

    typename std::vector<Vec<RING> >::const_iterator colIterEnd;
    if (B.getCols() < getCols())
    {
        colIterEnd = colIter + B.getCols();
    }
    else
    {
        colIterEnd = (m_entries.end());
    }

    for (; colIter < colIterEnd; ++colIter, ++BcolIter)
    {
        int comp = colIter->compare(*BcolIter);

        if (comp == -1)
        {
            return true;
        }
        if (comp == 1)
        {
            return false;
        }
    }

    if (B.getCols() <= getCols())
    {
        return false;
    }
    return true;
}

template<typename RING>
std::vector<unsigned int> Matrix<RING>::heightSequence(unsigned int start_row) const
{
    std::vector<unsigned int> res(m_n, 2);
    std::vector<unsigned int>::iterator resIter(res.begin());
    typename std::vector<Vec<RING> >::const_iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(m_entries.end());

    for (; colIter < colIterEnd; ++colIter, ++resIter)
    {
        *resIter = (colIter->height(start_row)).first;
    }
    return res;
}




template<typename RING>
void Matrix<RING>::swapRows(unsigned int row1, unsigned int row2)
{
    if(row1 != row2)
    {
        typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
        typename std::vector<Vec<RING> >::const_iterator colIterEnd(m_entries.end());

        for (; colIter < colIterEnd; ++colIter)
        {
            colIter->swap(row1, row2);
        }
    }
}

template<typename RING>
void Matrix<RING>::swapCols(unsigned int col1, unsigned int col2)
{
    std::swap(m_entries[col1], m_entries[col2]);
}

template<typename RING>
void Matrix<RING>::permuteCols(const Permutation& perm)
{
    perm.permute(m_entries.begin(), m_entries.end());
}

template<typename RING>
void Matrix<RING>::gaussElStep(unsigned int pivotRow, unsigned int pivotCol, bool doUpperEl, Matrix<RING>* const A)
{
    typename std::vector<Vec<RING> >::iterator pivotColIter(m_entries.begin() + pivotCol);
    std::pair<unsigned int, unsigned int> pivotSearchResult;
    pivotSearchResult = pivotColIter->height(pivotRow);
    swapRows(pivotRow, pivotSearchResult.second);

    unsigned int start_elPos(0);
    if (!doUpperEl)
    {
        start_elPos = pivotRow;
    }

    typename std::vector<Element<RING> >::iterator pivotColEntIter(pivotColIter->begin() + start_elPos);
    typename std::vector<Element<RING> >::const_iterator pivotColEntIterEnd(pivotColIter->end());
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(m_entries.end());

    std::vector<Element<RING> > factors(m_k - start_elPos, 1);
    typename std::vector<Element<RING> >::iterator factorsIterBeg(factors.begin());
    typename std::vector<Element<RING> >::const_iterator factorsIterEnd(factors.end());
    typename std::vector<Element<RING> >::iterator factorsIter(factorsIterBeg);
    typename std::vector<Element<RING> >::iterator pivotPosIter(factorsIterBeg + (pivotRow - start_elPos));

    Element<RING> pivotRowFactor = 1;
    for (; pivotColEntIter < pivotColEntIterEnd; ++pivotColEntIter, ++factorsIter)
    {
        if (factorsIter == pivotPosIter)
        {
            if (pivotSearchResult.first == 0) // a unit element
            {
                pivotRowFactor = pivotColEntIter->multInverse();
            }
        }
        else
        {
            if (pivotSearchResult.first == 0) // a unit element
            {
                *factorsIter = -(*pivotColEntIter);
            }
            else
            {
                if (pivotColEntIter->isZero() || pivotColEntIter->isOne())
                {
                    *factorsIter = 0;
                }
            }
        }
    }

    for (; colIter < colIterEnd; ++colIter)
    {
        Element<RING> pivot_rowEnt = colIter->get(pivotRow) * pivotRowFactor;
        typename std::vector<Element<RING> >::iterator colEntIterEnd(colIter->end());
        typename std::vector<Element<RING> >::iterator colEntIter(colIter->begin() + start_elPos);
        factorsIter = factorsIterBeg;

        for (; colEntIter < colEntIterEnd; ++colEntIter, ++factorsIter)
        {
            if (factorsIter == pivotPosIter)
            {
                *colEntIter = pivot_rowEnt;
            }
            else
            {
                *colEntIter += (*factorsIter) * pivot_rowEnt;
            }
        }
    }

    if (A != NULL)
    {
        A->swapRows(pivotRow, pivotSearchResult.second);
        typename std::vector<Vec<RING> >::iterator AcolIter(A->m_entries.begin());
        typename std::vector<Vec<RING> >::const_iterator AcolIterEnd(A->m_entries.end());
        for (; AcolIter < AcolIterEnd; ++AcolIter)
        {
            Element<RING> pivot_rowEnt = AcolIter->get(pivotRow) * pivotRowFactor;
            typename std::vector<Element<RING> >::iterator AcolEntIterEnd(AcolIter->end());
            typename std::vector<Element<RING> >::iterator AcolEntIter(AcolIter->begin() + start_elPos);
            factorsIter = factorsIterBeg;

            for (; AcolEntIter < AcolEntIterEnd; ++AcolEntIter, ++factorsIter)
            {
                if (factorsIter == pivotPosIter)
                {
                    *AcolEntIter = pivot_rowEnt;
                }
                else
                {
                    *AcolEntIter += (*factorsIter) * pivot_rowEnt;
                }
            }
        }
    }
}

template<typename RING>
Matrix<RING> Matrix<RING>::inverse(void) const
{
    Matrix<RING> copy(*this), res;
    res.identity(m_k);

    for (unsigned int col = 0; col < m_k; ++col)
    {
        copy.gaussElStep(col, col, true, &res);
    }
    return res;
}

template<typename RING>
Matrix<RING> Matrix<RING>::operator*(const Matrix<RING>& B) const
{
    Matrix<RING> res(this->getRows(), B.getCols());
    Matrix<RING> trans = this->transpose();

    for (unsigned int i = 0; i < this->getRows(); ++i)
    {
        const Vec<RING>& this_col = trans.getColumn(i);
        for (unsigned int j = 0; j < B.getCols(); j++)
        {
            res.m_entries[j].set(i, this_col.scalar(B.getColumn(j)));
        }
    }
    return res;
}

template<typename RING>
Vec<RING> Matrix<RING>::operator*(const Vec<RING>& b) const
{
    Vec<RING> res(getRows(), 0);
    typename std::vector<Vec<RING> >::const_iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd(m_entries.end());
    typename Vec<RING>::const_iterator bIter(b.begin());

    for (; colIter < colIterEnd; ++colIter, ++bIter)
    {
        res += (*colIter) * (*bIter);
    }
    return res;
}

template<typename RING>
std::vector<unsigned int> Matrix<RING>::toOrderedBasis(bool doUpperEl, Permutation* const trans, Matrix<RING>* const A)
{
    //    std::vector<unsigned int> blocksizes(RING::chainlength(), 0);
    RING::chainlength();
    std::vector<unsigned int> shape(RING::chainlength());
    if (A != NULL)
    {
        A->identity(m_k);
    }

    unsigned int pivot_col = 0, pivot_row = 0;
    bool isUnit = true;

    std::vector<int> pivotColList;

    while (true)
    {
        std::pair<unsigned int, unsigned int> res(m_entries[pivot_col].height(pivot_row));
        if ((isUnit && res.first == 0) || (!isUnit && res.first != RING::chainlength()))
        {
            gaussElStep(pivot_row, pivot_col, doUpperEl, A);
            ++pivot_row;
            pivotColList.push_back(pivot_col);
        }
        if (pivot_row == m_k)
        {
            if (isUnit)
            {
                shape[0] = m_k;
                if(RING::chainlength()==1)
                {
                    shape[1] = 0;
                }
            }
            else
            {
                shape[1] = m_k - shape[0];
            }
            break; // we have found an identity matrix, the matrix generates a free module
        }

        ++pivot_col;
        if (pivot_col == m_n)
        {
            if (isUnit)
            {
                isUnit = false;
                shape[0]  = pivot_row;
                pivot_col = 0;
                if(RING::chainlength()==1)
                {
                    break;
                }
            }
            else
            {
                shape[1] =  pivot_row - shape[0];
                break;
            }
        }
    }

    if (trans != NULL)
    {
        std::vector<int> permlist(getCols());
        std::copy(pivotColList.begin(), pivotColList.end(), permlist.begin());
        std::sort(pivotColList.begin(), pivotColList.end());
        std::vector<int>::iterator permlistIter = permlist.begin() + pivotColList.size();

        pivotColList.push_back(m_n); // dummy value
        std::vector<int>::const_iterator pivotIter = pivotColList.begin();
        for (int col = 0; col < (int) m_n; ++col)
        {
            if (col == *pivotIter)
            {
                ++pivotIter;
            }
            else
            {
                *permlistIter = col;
                permlistIter++;
            }
        }

        *trans = Permutation(permlist);
    }

    return shape;
}

template<typename RING>
std::vector<unsigned int> Matrix<RING>::ColumnSpan()
{
    std::vector<unsigned int> shape(RING::chainlength());
    std::set<unsigned int> pivotRows;
    std::set<unsigned int>::const_iterator pivotRowsIterEnd(pivotRows.end());
    typename std::vector<Vec<RING> >::iterator entriesIter(m_entries.begin()), pivotSwapPos(m_entries.begin()), iterEnd(m_entries.end());

    for(unsigned int desiredheight=0; desiredheight<RING::chainlength(); ++desiredheight)
    {
        entriesIter = pivotSwapPos;
        for(; entriesIter!=iterEnd; ++entriesIter)
        {
            if(pivotRows.size()==getRows())
            {
                return shape;
            }
            typename Vec<RING>::iterator rowIter(entriesIter->begin()), rowIterEnd(entriesIter->end());
            unsigned int r=0;
            for(;rowIter<rowIterEnd;++rowIter, ++r)
            {
                if(rowIter->height() == desiredheight && pivotRows.find(r) == pivotRowsIterEnd)
                {
                    Element<RING> pivotColFactor(1);
                    if(desiredheight==0)
                    {
                        pivotColFactor= rowIter->multInverse();
                        if(!pivotColFactor.isOne())
                        {
                            for(rowIter=entriesIter->begin();rowIter<rowIterEnd;++rowIter)
                            {//normalize the column
                                *rowIter *= pivotColFactor;
                            }
                        }
                    }

                    pivotRows.insert(r);
                    pivotRowsIterEnd = pivotRows.end();
                    ++shape[desiredheight];
                    if( entriesIter != pivotSwapPos)
                    {
                        std::swap(*entriesIter, *pivotSwapPos);
                    }

                    // gaussianElimination
                    typename std::vector<Vec<RING> >::iterator  gaussIter(pivotSwapPos+1);
                    typename Vec<RING>::const_iterator pivotRowIterBeg(pivotSwapPos->begin()), pivotRowIterEnd(pivotSwapPos->end());

                    for(; gaussIter<iterEnd; ++gaussIter)
                    {
                        Element<RING> loc_factor = - *(gaussIter->begin() +r);
                        if(!loc_factor.isZero())
                        {
                            if(desiredheight==1)
                            {
                                loc_factor = 1;
                            }
                            typename Vec<RING>::const_iterator pivotRowIter(pivotRowIterBeg);
                            typename Vec<RING>::iterator gaussRowIter(gaussIter->begin());
                            for( ; pivotRowIter<pivotRowIterEnd; ++pivotRowIter, ++gaussRowIter)
                            {
                                *gaussRowIter+= loc_factor* (*pivotRowIter);
                            }
                        }
                    }
                    ++pivotSwapPos;
                    break; //rowIter
                }
            }
        }
    }

    return shape;
}



// Span(G1) contains Span(G2), we want to have a basis of W (as matrix again) such that  Span(G1) =  Span(G2) \oplus W
template<>
unsigned int Matrix<Field>::complement(const Matrix<Field>& G1, const Matrix<Field>& G2, Matrix<Field>& res);

template<typename RING>
void Matrix<RING>::VerticalJoin(const Matrix<RING>& G1, const Matrix<RING>& G2, Matrix<RING>& res)
{
    res = Matrix<RING>(G1.getRows()+G2.getRows(), G1.getCols() );

    typename std::vector<Vec<RING> >::const_iterator G1Iter(G1.getEntries().begin());
    typename std::vector<Vec<RING> >::const_iterator  G1IterEnd(G1.getEntries().end());
    typename std::vector<Vec<RING> >::const_iterator  G2Iter(G2.getEntries().begin());
    typename std::vector<Vec<RING> >::iterator  resIter(res.m_entries.begin());

    for(; G1Iter<G1IterEnd; ++G1Iter, ++G2Iter, ++resIter)
    {
        typename Vec<RING>::iterator col_resIter(resIter->begin());

        typename Vec<RING>::const_iterator col1Iter(G1Iter->begin());
        typename Vec<RING>::const_iterator col1IterEnd(G1Iter->end());
        for(; col1Iter<col1IterEnd; ++col1Iter, ++col_resIter)
        {
            *col_resIter = *col1Iter;
        }

        typename Vec<RING>::const_iterator col2Iter(G2Iter->begin());
        typename Vec<RING>::const_iterator col2IterEnd(G2Iter->end());
        for(; col2Iter<col2IterEnd; ++col2Iter, ++col_resIter)
        {
            *col_resIter = *col2Iter;
        }
    }
}

template<typename RING>
void Matrix<RING>::rowSubmatrix(const Matrix<RING>& G1, unsigned int begin, unsigned int nr_rows)
{
    *this = Matrix<RING>(nr_rows, G1.getCols());
    typename std::vector<Vec<RING> >::const_iterator G1Iter(G1.getEntries().begin());
    typename std::vector<Vec<RING> >::const_iterator  G1IterEnd(G1.getEntries().end());
    typename std::vector<Vec<RING> >::iterator  resIter(m_entries.begin());

    for(; G1Iter<G1IterEnd; ++G1Iter, ++resIter)
    {
        typename Vec<RING>::iterator col_resIter(resIter->begin());

        typename Vec<RING>::const_iterator col1Iter(G1Iter->begin()+begin);
        typename Vec<RING>::const_iterator col1IterEnd(col1Iter + nr_rows);
        for(; col1Iter<col1IterEnd; ++col1Iter, ++col_resIter)
        {
            *col_resIter = *col1Iter;
        }
    }
}

template<typename RING>
void Matrix<RING>::columnSubmatrix(unsigned int new_n)
{
    if(new_n < m_n)
    {
        m_entries.resize(new_n);
        m_n = new_n;
    }
}

template<typename RING>
Matrix<RING> Matrix<RING>::transpose(void) const
{
    Matrix<RING> t(m_n, m_k);

    for (unsigned int i = 0; i < m_k; ++i)
    {
        for (unsigned int j = 0; j < m_n; ++j)
        {
            t.m_entries[i].set(j, this->getElement(i, j));
        }
    }
    return t;
}

template<typename RING>
void Matrix<RING>::addRows(unsigned int fix, unsigned int to)
{
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());

    for (; colIter < colIterEnd; ++colIter)
    {
        colIter->operator[](to) += colIter->operator[](fix);
    }
}

template<typename RING>
void Matrix<RING>::multiplyColumn(unsigned int col, const Element<RING>& a)
{
    m_entries[col] *= a;
}

template<typename RING>
void Matrix<RING>::normalizeColumns(AlgTypeE algtype)
{
    if(algtype==PERMUTATIONAL)
    {
        return;
    }
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());

    for (; colIter < colIterEnd; ++colIter)
    {
        colIter->normalize(algtype==PLUSMINUSMONOMIAL);
    }
}

template<typename RING>
void Matrix<RING>::multiplyRow(unsigned int row, const Element<RING>& a)
{
    if (a.isOne())
    {
        return;
    }
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());

    for (; colIter < colIterEnd; ++colIter)
    {
        colIter->operator[](row) *= a;
    }
}

template<typename RING>
void Matrix<RING>::multiplyRows(const std::vector<unsigned int>& rows, const Element<RING>& a, const std::vector<
        unsigned int>& notTouchCol)
        {
    if (a.isOne())
    {
        return;
    }

    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());
    std::vector<unsigned int>::const_iterator rowIterBeg(rows.begin());
    std::vector<unsigned int>::const_iterator rowIterEnd(rows.end());
    std::vector<unsigned int>::const_iterator rowIter;

    std::vector<unsigned int>::const_iterator notTouchIter(notTouchCol.begin()), notTouchIterEnd(notTouchCol.end());
    unsigned int col = 0;

    for (; colIter < colIterEnd; ++colIter, ++col)
    {
        if (notTouchIter != notTouchIterEnd && col == *notTouchIter)
        {
            ++notTouchIter;
            continue;
        }
        for (rowIter = rowIterBeg; rowIter < rowIterEnd; ++rowIter)
        {
            colIter->multEntry(*rowIter, a);
        }
    }
        }

template<typename RING>
void Matrix<RING>::replaceLastRow(const Vec<RING>& new_row)
{
    typename Vec<RING>::const_iterator new_rowIter(new_row.begin());
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());
    for (; colIter < colIterEnd; ++colIter, ++new_rowIter)
    {
        colIter->back() = *new_rowIter;
    }
}

template<typename RING>
void Matrix<RING>::appendZeroRow(void)
{
    ++m_k;
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());
    for (; colIter < colIterEnd; ++colIter)
    {
        colIter->push_back(0);
    }
}

template<typename RING>
void Matrix<RING>::removeLastRow(void)
{
    --m_k;
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());
    for (; colIter < colIterEnd; ++colIter)
    {
        colIter->pop_back();
    }
}


template<typename RING>
void Matrix<RING>::appendColumn(const Vec<RING>& newCol)
{
    if (m_n == 0)
    {
        m_k = newCol.size();
    }
    ++m_n;
    m_entries.push_back(newCol);
}

template<typename RING>
void Matrix<RING>::appendColumns(const std::vector<Vec<RING> >& newCols)
{
    if(newCols.size()!=0)
    {
        m_entries.resize(m_entries.size()+newCols.size());
        std::copy(newCols.begin(), newCols.end(), m_entries.begin()+m_n);

        if (m_n == 0)
        {
            m_k = newCols[0].size();
        }
        m_n +=  newCols.size();
    }
}

template<typename RING>
void Matrix<RING>::replaceLastColumn(const Vec<RING>& newCol)
{
    m_entries.back() = newCol;
}

template<typename RING>
void Matrix<RING>::removeLastColumn(void)
{
    m_n--;
    m_entries.pop_back();
}


template<typename RING>
void Matrix<RING>::identity(unsigned int k)
{
    m_k = k;
    m_n = k;
    m_entries = std::vector<Vec<RING> >(m_k, Vec<RING> (m_k));
    for (unsigned int i = 0; i < k; ++i)
    {
        m_entries[i].set(i, 1);
    }
}

template<typename RING>
void Matrix<RING>::systematicDiagonal(unsigned int k0, unsigned int k1)
{
    m_k = k0 + k1;
    m_n = m_k;
    m_entries = std::vector<Vec<RING> >(m_k, Vec<RING> (m_k));
    for (unsigned int i = 0; i < k0; ++i)
    {
        m_entries[i][i] = 1;
    }
    for (unsigned int i = k0; i < m_k; ++i)
    {
        m_entries[i][i] = 2;
    }
}

template<typename RING>
void Matrix<RING>::random(unsigned int chainlength, unsigned int* type, unsigned int n)
{
    m_k = type[0];
    for (unsigned int i = 1; i < chainlength; ++i)
    {
        m_k += type[i];
    }
    if (n < m_k)
    {
        std::cerr << "Cannot initialize a random matrix of this type, because the given length was to short"
                << std::endl;
        exit(-1);
    }

    m_n = n;
    m_entries = std::vector<Vec<RING> >(n, Vec<RING> (m_k, 0));

    unsigned int startpos(0);
    for (unsigned int t = 0; t < chainlength; ++t)
    {
        unsigned int endpos = startpos + type[t];
        const Element<RING>& thetapow = RING::thetapow(t);
        for (unsigned int i = startpos; i < endpos; ++i)
        {
            m_entries[i].set(i, thetapow);
            for (unsigned int j = m_k; j < n; ++j)
            {
                m_entries[j].set(i, thetapow * RING::random());
            }
        }
        startpos = endpos;
    }
}

template<typename RING>
void Matrix<RING>::minimizeByColMult(bool only_minusOne, unsigned int row, UnordPartition& row_partition,
        const std::vector<unsigned int>& fixedCols, Matrix<RING>* transporterMat, Vec<RING>* transporterMult)
        {
    int join_setIndex(-1);
    unsigned int col(fixedCols.back());
    std::vector<unsigned int> fixedCopy(fixedCols.begin(), fixedCols.end());
    fixedCopy.pop_back();
    std::sort(fixedCopy.begin(), fixedCopy.end());
    Element<RING> minusOne( -Element<RING>(1));

    while (row > 0)
    {
        row--;
        Element<RING> el = getElement(row, col);
        if (el.height() > 0)
        { // a non unit element
            continue;
        }

        Element<RING> block_factor;
        Element<RING> block_factor_inverse;
        if(only_minusOne)
        {
            if( el > -el )
            {
                block_factor = minusOne;
                block_factor_inverse = minusOne;
            }
            else
            {
                block_factor = Element<RING>(1);
                block_factor_inverse = block_factor;
            }
        }
        else
        {
            block_factor = el.multInverse();
            block_factor_inverse = el;
        }

        int setIndex = row_partition.m_index[row];
        if (join_setIndex == -1)
        {
            join_setIndex = setIndex;
            if (!block_factor.isOne())
            {
                multiplyColumn(col, block_factor);
                if (transporterMult != NULL)
                {
                    transporterMult->multEntry(col, block_factor);
                }
            }
        }
        else
        {
            if (join_setIndex != setIndex)
            {
                if (!block_factor.isOne())
                {
                    //	colMult(setindex, el, NULL);
                    std::vector<unsigned int> rows;
                    row_partition.getRows(setIndex, rows);

                    multiplyRows(rows, block_factor, fixedCopy);
                    if (transporterMat != NULL)
                    {
                        transporterMat->multiplyRows(rows, block_factor);
                        std::vector<unsigned int>::const_iterator fixIter = fixedCols.begin();
                        std::vector<unsigned int>::const_iterator fixIterEnd = fixIter + (fixedCols.size() - 1);
                        std::vector<unsigned int>::const_iterator rowsBeg = rows.begin();
                        std::vector<unsigned int>::const_iterator rowsEnd = rows.end();
                        for (; fixIter < fixIterEnd; ++fixIter)
                        {
                            std::pair<unsigned int, unsigned int> hRes = getColumn(*fixIter).height();
                            if (hRes.first == 0)
                            {
                                for (std::vector<unsigned int>::const_iterator rowsIter(rowsBeg); rowsIter < rowsEnd; ++rowsIter)
                                {
                                    if (*rowsIter == hRes.second)
                                    {
                                        transporterMult->multEntry(*fixIter, block_factor_inverse);
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
                row_partition.unify(join_setIndex, setIndex);
            }
        }
    }
        }



template<>
void Matrix<Field>::applyFrobenius(unsigned int frob_pow, const std::vector<bool>* const isfixedCol);

template<typename RING>
unsigned int Matrix<RING>::checkMatrix(const Matrix<RING>& Gamma, Matrix<RING>& Delta)
{
    Matrix<RING> G(Gamma); // this also allows us to pass Delta == *this and overwrite *this
    Permutation trans;
    unsigned int n = G.getCols();
    unsigned int k = G.getRows();
    std::vector<unsigned int> blocksize(G.toOrderedBasis(true, &trans, NULL));
    unsigned int k0 = blocksize[0]; // trans*G is a systematic matrix
    unsigned int k1 =0;
    if(RING::chainlength()==2)
    {
        k1 = blocksize[1];
    }
    if(k0+k1 != k)
    {
        std::cerr<<"The matrix does not have full row rank"<<std::endl;
    }

    Permutation transInv;
    Permutation::inv(trans, transInv);
    G.permuteCols(transInv);
    Delta = Matrix<RING> (n - k0, n);

    // G is now a block matrix of the following form
    // I_k0     A                   B
    //   0     \theta*I_k1  \theta*C
    // Delta can be calculated by
    // (A*C - B)^T            -C^T           I_n-k0-k1
    // -\theta*A^T          \theta*I_k1            0

    // we start with the identity matrix
    for (unsigned int col = k; col < n; ++col)
    {
        Delta.m_entries[col].set(col - k, 1);
    }

    // now the rest,depending on the type of the code
    if (k1 == 0)
    {
        // there is only the B
        for (unsigned int col = 0; col < k0; ++col)
        {
            for (unsigned int row = 0; row < n - k; ++row)
            {
                Delta.m_entries[col].set(row, -G.getElement(col, k + row));
            }
        }
    }
    else
    {
        // we start with filling A and C
        Element<RING> theta = RING::thetapow(1);
        Element<RING> minusTheta = -RING::thetapow(1);
        Matrix<RING> A(k0, k1);
        for (unsigned int col = 0; col < k1; ++col)
        {
            for (unsigned int row = 0; row < k0; ++row)
            {
                A.m_entries[col].set(row, G.getElement(row, k0 + col));
                Delta.m_entries[row].set(col + n - k, minusTheta * A.getElement(row, col));
            }
        }

        Matrix<RING> C(k1, n - k);
        Element<RING> minusOne = -Element<RING> (1);
        for (unsigned int col = 0; col < n - k; ++col)
        {
            for (unsigned int row = 0; row < k1; ++row)
            {
                if (!G.getElement(k0 + row, k + col).isZero())
                {
                    C.m_entries[col].set(row, 1);
                    Delta.m_entries[k0 + row].set(col, minusOne);
                }
            }
        }

        // \theta*I_k1
        for (unsigned int i = 0; i < k1; ++i)
        {
            Delta.m_entries[k0 + i].set(n - k + i, theta);
        }

        // the upper left block
        Matrix<RING> AC(A * C);
        for (unsigned int col = 0; col < k0; ++col)
        {
            for (unsigned int row = 0; row < n - k; ++row)
            {
                Delta.m_entries[col].set(row, AC.getElement(col, row) - G.getElement(col, k + row));
            }
        }
    }

    //        std::cout<<"In get Check"<<std::endl;
    //        std::cout<<G<<std::endl;
    //        std::cout<<Delta<<std::endl;
    //        std::cout<<G*Delta.transpose()<<std::endl;

    Delta.permuteCols(trans);
    return k0;
}



template<typename RING>
Matrix<RING>::Matrix(unsigned int k, unsigned int n, const DenseVec<BinaryField>& m) :
m_n(n), m_k(k), m_entries(std::vector<Vec<RING> >(n, Vec<RING> (k, 0)))
{
    typename std::vector<Vec<RING> >::iterator colIter(m_entries.begin());
    typename std::vector<Vec<RING> >::iterator colIterEnd(m_entries.end());

    unsigned int bitposBegin = 0;
    unsigned int bitposEnd = RING::bitsPerEntry();
    unsigned int readpos = 0;

    uchar mask((((uchar) 1) << RING::bitsPerEntry()) - 1);

    for (; colIter < colIterEnd; ++colIter)
    {
        typename Vec<RING>::iterator colEntIter(colIter->begin());
        typename Vec<RING>::iterator colEntIterEnd(colIter->end());

        for (; colEntIter < colEntIterEnd; ++colEntIter)
        {
            uchar entry;

            if (bitposEnd <= NBITS)
            {
                entry = (m.m_data[readpos] >> bitposBegin) & mask;
                bitposBegin = bitposEnd;
                bitposEnd += RING::bitsPerEntry();
            }
            else
            {
                if (bitposBegin != NBITS)
                {
                    entry = (m.m_data[readpos] >> bitposBegin) & mask;
                }
                else
                {
                    entry = 0;
                }
                ++readpos;
                entry |= (m.m_data[readpos] & ((((uchar) 1) << (bitposEnd - NBITS)) - 1)) << (NBITS - bitposBegin);
                bitposBegin = bitposEnd - NBITS;
                bitposEnd = bitposBegin + RING::bitsPerEntry();
            }
            *colEntIter = entry;
        }
    }
}
/*
template<typename RING>
unsigned int Matrix<RING>::getDualMinDist(const unsigned int* const type, const KrawtchoukTable& krawtchouk) const
{
    unsigned int k(type[0]);
    unsigned int n(getCols());
    if(RING::chainlength() == 2)
    {
        k+=type[1];
    }

    if( type[0]<= n-k ) // primal is easier
    {
        Distribution<typename RING::MyWeight> primal_wd(*this);
        return krawtchouk.getDualMinDist(primal_wd);
    }

    if(type[0] == n)
    {
        return std::numeric_limits<unsigned int>::max();
    }

    Matrix<RING> Delta;
    Matrix<RING>::checkMatrix(*this, Delta);
    return Delta.calcMinDist(0);
}

template<typename RING>
unsigned int Matrix<RING>::getMinDist(const unsigned int* const type, const KrawtchoukTable& krawtchouk) const
{
    unsigned int k(type[0]);
    unsigned int n(getCols());
    if(RING::chainlength() == 2)
    {
        k+=type[1];
    }

    if( type[0] <= n - k ) // primal is easier
    {
        if(k==0)
        {
            return std::numeric_limits<unsigned int>::max();
        }
        return this->calcMinDist(0);
    }

    Matrix<RING> Delta;
    Matrix<RING>::checkMatrix(*this, Delta);
    Distribution<typename RING::MyWeight> dual_wd(Delta);
    return  krawtchouk.getDualMinDist(dual_wd);
}*/

template<typename RING>
unsigned int Matrix<RING>::weightOfLastRow(void) const
{
    typename RING::MyWeight w;
    typename std::vector<Vec<RING> >::const_iterator colIter( m_entries.begin());
    typename std::vector<Vec<RING> >::const_iterator colIterEnd( m_entries.end() );
    for( ; colIter<colIterEnd; ++colIter)
    {
        w.add(colIter->back().height());
    }
    return w.HammingWeight();
}
/*
template<typename RING>
bool Matrix<RING>::verifyMinDist(const unsigned int* const type, const KrawtchoukTable& krawtchouk, unsigned int d, unsigned int desired_dual_d, bool residual) const
{
    unsigned int k(type[0]);
    unsigned int n(getCols());
    if(RING::chainlength() == 2)
    {
        k+=type[1];
    }

    if( type[0]<= n-k ) // primal is easier
    {
        Distribution<typename RING::MyWeight> primal_wd(*this);
        if(residual && primal_wd.getMinWeight() < d)
        {
            return false;
        }

        return ( krawtchouk.getDualMinDist(primal_wd)>=desired_dual_d );
    }

    if(type[0] == n)
    {
        return true;
    }

    Matrix<RING> Delta;
    Matrix<RING>::checkMatrix(*this, Delta);
    Distribution<typename RING::MyWeight> dual_wd(Delta);
    if(dual_wd.getMinWeight() < desired_dual_d)
    {
        return false;
    }

    if(residual && krawtchouk.getDualMinDist(dual_wd) < d)
    {
        return false;
    }
    return true;
}*/

template<typename RING>
unsigned int Matrix<RING>::calcMinDist(unsigned int desired_d) const
{
    Matrix<RING> copy( *this );
    std::vector<unsigned int> type = copy.toOrderedBasis(false, NULL, NULL);
    unsigned int d(std::numeric_limits<unsigned int>::max()), w;
    if(RING::chainlength() > 1)
    {
        type[1] = copy.getRows() - type[0];
    }
    CodewordEnumerator<RING> enumerator(type, copy);
    while(++enumerator) // skips the all zero word
    {
        w = enumerator->grayImageWeight();
        if( d>w )
        {
            d = w;
            if(d<desired_d)
            {
                return 0;
            }
        }
    };

    return d;
}

template<typename RING>
unsigned int Matrix<RING>::fillLastColProjective(unsigned int rank)
{
    Vec<RING>& last_Col( m_entries.back());
    return last_Col.fill(rank, true);
}

#endif /*MATRIX_HPP_*/
