/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  vector.cpp
///  \brief Implements specialization of vectors.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   07.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "vector.hpp"
#include "field.hpp"
#include "z4.hpp"
#include "f2_f2.hpp"

char DenseVec<BinaryField>::bits_in_16bits[0x1u << 16];

int DenseVec<BinaryField>::iterated_bitcount(uchar n)
        {
    n = (n & 0x5555555555555555ll) + ((n >> 1) & 0x5555555555555555ll);
    n = (n & 0x3333333333333333ll) + ((n >> 2) & 0x3333333333333333ll);
    n = (n & 0x0f0f0f0f0f0f0f0fll) + ((n >> 4) & 0x0f0f0f0f0f0f0f0fll);
    n = (n & 0x00ff00ff00ff00ffll) + ((n >> 8) & 0x00ff00ff00ff00ffll);
    n = (n & 0x0000ffff0000ffffll) + ((n >> 16) & 0x0000ffff0000ffffll);
    n = (n & 0x00000000ffffffffll) + ((n >> 32) & 0x00000000ffffffffll);
    return (int) n;
        }

void DenseVec<BinaryField>::compute_bits_in_16bits(void)
        {
    uchar i;
    for (i = 0; i < (0x1u << 16); ++i)
        bits_in_16bits[i] = iterated_bitcount(i);
    return;
        }

int DenseVec<BinaryField>::bit_count(uchar n)
        {
    return bits_in_16bits[n & 0xffffu] + bits_in_16bits[(n >> 16) & 0xffffu] + bits_in_16bits[(n >> 32) & 0xffffu]
                                                                                              + bits_in_16bits[(n >> 48) & 0xffffu];
        }

//std::ostream& operator<<(std::ostream& os, const DenseVec<BinaryField>& v)
//{
//    DenseVec<BinaryField>::Iterator iter(v.getIterator());
//    do
//    {
//        os<<((int) *iter);
//    }while(++iter);
//}


template<>
void Vec<Field>::applyFrobenius(unsigned int frob_pow)
{
    if (frob_pow == 0)
    {
        return;
    }

    const std::vector<unsigned int>& img = Field::getFrobenius(frob_pow);
    Vec<Field>::iterator iter(this->begin());
    Vec<Field>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter)
    {
        *iter = Element<Field> (img[iter->getIndex()]);
    }
}

template<>
DenseVec<Field> DenseVec<Field>::operator*(const Element<Field>& x) const
{
    return Vec<Field>::operator*(x);
}

template<>
SymWeight<1u> DenseVec<Field>::getWeight(void) const
{
    return SymWeight<1u> (grayImageWeight());
}

template<>
unsigned int DenseVec<Field>::grayImageWeight(void) const
{
    unsigned int w(0);
    DenseVec<Field>::const_iterator iter(this->begin());
    DenseVec<Field>::const_iterator iterEnd(this->end());

    for (; iter < iterEnd; ++iter)
    {
        if (!iter->isZero())
        {
            ++w;
        }
    }
    return w;
}

DenseVec<BinaryField>::DenseVec(int i) :
            m_len(i), m_data((i - 1) / 64 + 1, 0)
            {
            }

DenseVec<BinaryField>::DenseVec(const std::vector<Element<BinaryField> >& v) :
            m_len(v.size())
            {
    std::vector<Element<BinaryField> >::const_iterator iter(v.begin());
    std::vector<Element<BinaryField> >::const_iterator iterEnd(v.end());
    unsigned int shift(0);
    uchar word(0);
    for (; iter < iterEnd; ++iter)
    {
        word += ((uchar) iter->getIndex()) << shift;
        ++shift;
        if (shift == 64)
        {
            m_data.push_back(word);
            shift = 0;
            word = 0;
        }
    }
    if (shift != 0)
    {
        m_data.push_back(word);
    }
            }

SymWeight<1u> DenseVec<BinaryField>::getWeight(const DenseVec& mask) const
        {
    SymWeight<1u> w;
    std::vector<uchar >::const_iterator iterVec(m_data.begin());
    std::vector<uchar >::const_iterator iterVecEnd(m_data.end());
    std::vector<uchar >::const_iterator maskIter(mask.m_data.begin());
    for (; iterVec < iterVecEnd; ++iterVec, ++maskIter)
    {
        w.qty += DenseVec<BinaryField>::bit_count(*iterVec & *maskIter);
    }
    return w;
        }

SymWeight<1u> DenseVec<BinaryField>::getWeight(void) const
        {
    return SymWeight<1u> (grayImageWeight());
        }

unsigned int DenseVec<BinaryField>::grayImageWeight(void) const
        {
    unsigned int w(0);
    std::vector<uchar >::const_iterator iterVec(m_data.begin());
    std::vector<uchar >::const_iterator iterVecEnd(m_data.end());
    for (; iterVec < iterVecEnd; ++iterVec)
    {
        w += DenseVec<BinaryField>::bit_count(*iterVec);
    }
    return w;
        }

DenseVec<BinaryField>& DenseVec<BinaryField>::operator+=(const DenseVec& b)
        {
    std::vector<uchar >::iterator iterVec(m_data.begin());
    std::vector<uchar >::iterator iterVecEnd(m_data.end());
    std::vector<uchar >::const_iterator bIter(b.m_data.begin());
    for (; iterVec < iterVecEnd; ++iterVec, ++bIter)
    {
        *iterVec ^= *bIter;
    }
    return *this;
        }

DenseVec<BinaryField>& DenseVec<BinaryField>::operator |=(const DenseVec& b)
        {
    std::vector<uchar >::iterator iterVec(m_data.begin());
    std::vector<uchar >::iterator iterVecEnd(m_data.end());
    std::vector<uchar >::const_iterator bIter(b.m_data.begin());
    for (; iterVec < iterVecEnd; ++iterVec, ++bIter)
    {
        *iterVec |= *bIter;
    }
    return *this;
        }

DenseVec<BinaryField>& DenseVec<BinaryField>::operator&=(const DenseVec& b)
                {
    std::vector<uchar >::iterator iterVec(m_data.begin());
    std::vector<uchar >::iterator iterVecEnd(m_data.end());
    std::vector<uchar >::const_iterator bIter(b.m_data.begin());
    for (; iterVec < iterVecEnd; ++iterVec, ++bIter)
    {
        *iterVec &= *bIter;
    }
    return *this;
                }

DenseVec<BinaryField> DenseVec<BinaryField>::operator~(void) const
                {
    DenseVec<BinaryField> res(*this);
    std::vector<uchar >::iterator iterVec(res.m_data.begin());
    std::vector<uchar >::iterator iterVecEnd(res.m_data.end());

    for (; iterVec < iterVecEnd; ++iterVec)
    {
        *iterVec = ~(*iterVec);
    }

    uchar& lastLimb = res.m_data.back();
    unsigned int first_nonValidIndex(m_len & BITPOSMASK);
    if( first_nonValidIndex != 0)
    {
        uchar cleanMask = (((uchar) 1)<<first_nonValidIndex) -1;
        lastLimb &= cleanMask;
    }
    return res;
                }

DenseVec<BinaryField> DenseVec<BinaryField>::operator<<(unsigned int shift)
{
//    if(shift>this->m_len)
//    {
//        std::cout<<"boese"<<std::endl;
//    }
//
//
//    std::cout<<"operator<< "<<shift<<" "<<this->m_len<<std::endl;

    DenseVec<BinaryField> res(this->m_len - shift);
//    std::cout<<res.m_len<<" "<<res.m_data.size()<<std::endl;
//    help::print(std::cout, res.m_data, " ")<<std::endl;
    unsigned int data_shift = shift>>6;
    unsigned int right_bit_shift = shift & BITPOSMASK;
    unsigned int left_bit_shift = 64 - right_bit_shift;
    uchar right_mask = ((uchar) 1)<<right_bit_shift;
    --right_mask;

    std::vector<uchar>::iterator resIter(res.m_data.begin());
    std::vector<uchar>::const_iterator vecIter(this->m_data.begin() + data_shift);
    std::vector<uchar>::const_iterator resIterEnd(--res.m_data.end());

    while(resIter != resIterEnd)
    {
        *resIter = (*vecIter)>>right_bit_shift;
        ++vecIter;
        *resIter |=  (*vecIter&right_mask)<<left_bit_shift;
        ++resIter;
    }
    *resIter = (*vecIter)>>right_bit_shift;
    ++vecIter;
    if(vecIter != this->m_data.end())
    {
        *resIter |=  (*vecIter&right_mask)<<left_bit_shift;
    }

    return res;
}

void DenseVec<BinaryField>::clear(void)
                {
    std::vector<uchar >::iterator iterVec(m_data.begin());
    std::vector<uchar >::iterator iterVecEnd(m_data.end());

    for (; iterVec < iterVecEnd; ++iterVec)
    {
        *iterVec = 0;
    }
                }


std::ostream& operator<<(std::ostream& os, const DenseVec<BinaryField>& c)
{
    std::vector<uchar >::const_iterator iterVec(c.m_data.begin());
    uchar word = *iterVec;
    for (unsigned int i = 0; i < c.m_len; ++i)
    {
        if ((i & DenseVec<BinaryField>::BITPOSMASK) == 0)
        {
            word = *iterVec;
            ++iterVec;
        }
        os << (word & 1) << " ";
        word >>= 1;
    }
    return os;
}

DenseVec<Z4>::DenseVec(int i) :
            m_len(i), m_data((i - 1) / (int) ENTRIES_PER_WORD + 1, 0)
            {
            }

DenseVec<Z4>::DenseVec(const std::vector<Element<Z4> >& v) :
            m_len(v.size())
            {
    uchar word(0);
    uchar shift(0);
    std::vector<Element<Z4> >::const_iterator iter(v.begin());
    std::vector<Element<Z4> >::const_iterator iterEnd(v.end());
    for (; iter < iterEnd; ++iter)
    {
        word += ((uchar) iter->getIndex()) << (3 * shift);
        ++shift;
        if (shift == ENTRIES_PER_WORD)
        {
            m_data.push_back(word);
            shift = 0;
            word = 0;
        }
    }
    if (shift != 0)
    {
        m_data.push_back(word);
    }
            }

std::vector<Element<Z4> > DenseVec<Z4>::getVector(void) const
        {
    std::vector<Element<Z4> > res(m_len, 0);
    std::vector<uchar>::const_iterator iterData(m_data.begin());
    std::vector<Element<Z4> >::iterator iterVec(res.begin());
    std::vector<Element<Z4> >::iterator iterVecEnd(res.end());
    unsigned int wordpos = 0;
    uchar word(*iterData);

    for (; iterVec < iterVecEnd; ++iterVec)
    {
        *iterVec = word & 3;
        ++wordpos;
        if (wordpos == ENTRIES_PER_WORD)
        {
            wordpos = 0;
            ++iterData;
            word = *iterData;
        }
        else
        {
            word >>= 3;
        }
    }
    return res;
        }

void DenseVec<Z4>::getHeightSeq(std::vector<unsigned int>::iterator iterRes,
        std::vector<unsigned int>::iterator iterResEnd) const
        {
    std::vector<uchar>::const_iterator iterData(m_data.begin());
    unsigned int wordpos = 0;
    uchar word(*iterData);

    for (; iterRes != iterResEnd; ++iterRes)
    {
        if (word & 1)
        {
            *iterRes = 0;
        }
        else if (word & 2)
        {
            *iterRes = 1;
        }
        else
        {
            *iterRes = 2;
        }
        ++wordpos;
        if (wordpos == ENTRIES_PER_WORD)
        {
            wordpos = 0;
            ++iterData;
            word = *iterData;
        }
        else
        {
            word >>= 3;
        }
    }
        }

std::ostream& operator<<(std::ostream& os, const DenseVec<Z4>& c)
{
    Vec<Z4> v(c.getVector());
    os << v << std::endl;
    return os;
}

SymWeight<2u> DenseVec<Z4>::getWeight(void) const
        {
    SymWeight<2u> w;

    std::vector<uchar >::const_iterator iterVec(m_data.begin());
    std::vector<uchar >::const_iterator iterVecEnd(m_data.end());
    for (; iterVec < iterVecEnd; ++iterVec)
    {
        uchar myword(*iterVec & (*iterVec << 1)); // there is a one where the entry is equal to 3
        unsigned int nr3 = DenseVec<BinaryField>::bit_count(myword); // we have counted all 3's
        myword = *iterVec & ONEBITMASK; // there is a one where the entry is equal to 1 or 3
        unsigned int nr1_3 = DenseVec<BinaryField>::bit_count(myword); // we have counted all units
        unsigned int nr2 = DenseVec<BinaryField>::bit_count(*iterVec) - nr3 - nr1_3; // we have counted all 2's
        w.add(0, nr1_3);
        w.add(1, nr2);
    }
    return w;
        }

unsigned int DenseVec<Z4>::grayImageWeight(void) const
        {
    SymWeight<2u> w(getWeight());
    return w.qty[0] + ( w.qty[1]<<1);
        }

SymWeight<2u> DenseVec<Z4>::getWeight(const DenseVec<Z4>& mask) const
        {
    std::vector<unsigned int> cwe(getCompleteWeight(mask));
    SymWeight<2u> w;
    w.add(0, cwe[1] + cwe[3]);
    w.add(1, cwe[2]);
    return w;
        }

std::vector<unsigned int> DenseVec<Z4>::getCompleteWeight(const DenseVec<Z4>& mask) const
        {
    std::vector<unsigned int> res(4, 0);
    unsigned int& res1 = res[1];
    unsigned int& res2 = res[2];
    unsigned int& res3 = res[3];

    std::vector<uchar >::const_iterator maskIter(mask.m_data.begin());
    std::vector<uchar >::const_iterator iter(m_data.begin());
    std::vector<uchar >::const_iterator iterEnd(m_data.end());
    for (; iter < iterEnd; ++iter, ++maskIter)
    {
        uchar maskedWord = *iter & *maskIter;
        if (maskedWord) // the word is not zero
        {
            uchar myword(maskedWord & (maskedWord << 1)); // there is a one where the entry is equal to 3
            unsigned int nr3 = DenseVec<BinaryField>::bit_count(myword); // we have counted all 3's
            myword = maskedWord & ONEBITMASK; // there is a one where the entry is equal to 1 or 3
            unsigned int nr1 = DenseVec<BinaryField>::bit_count(myword) - nr3; // we have counted all units
            unsigned int nr2 = DenseVec<BinaryField>::bit_count(maskedWord) - ((nr3 << 1) + nr1); // we have counted all 2's
            res1 += nr1;
            res2 += nr2;
            res3 += nr3;
        }
    }
    return res;
        }

//std::vector<Element<F2_F2> > getVector(void) const;
DenseVec<F2_F2>::DenseVec(int i) :
            m_len(i), m_data(((i - 1) >> 5) + 1, 0)
            {
            }

DenseVec<F2_F2>::DenseVec(const std::vector<Element<F2_F2> >& v) :
            m_len(v.size())
            {
    uchar word(0);
    uchar shift(0);
    std::vector<Element<F2_F2> >::const_iterator iter(v.begin());
    std::vector<Element<F2_F2> >::const_iterator iterEnd(v.end());
    for (; iter < iterEnd; ++iter)
    {
        word |= ((uchar) iter->getIndex()) << shift;
        shift += 2;
        if (shift == 64)
        {
            m_data.push_back(word);
            shift = 0;
            word = 0;
        }
    }
    if (shift != 0)
    {
        m_data.push_back(word);
    }
            }

SymWeight<2u> DenseVec<F2_F2>::getWeight(const DenseVec<F2_F2>& mask) const
        {
    std::vector<unsigned int> cwe(getCompleteWeight(mask));
    SymWeight<2u> w;
    w.add(0, cwe[1] + cwe[3]);
    w.add(1, cwe[2]);
    return w;
        }

std::vector<unsigned int> DenseVec<F2_F2>::getCompleteWeight(const DenseVec<F2_F2>& mask) const
        {
    std::vector<unsigned int> res(4, 0);
    unsigned int& res1 = res[1];
    unsigned int& res2 = res[2];
    unsigned int& res3 = res[3];

    std::vector<uchar >::const_iterator maskIter(mask.m_data.begin());
    std::vector<uchar >::const_iterator iter(m_data.begin());
    std::vector<uchar >::const_iterator iterEnd(m_data.end());
    for (; iter < iterEnd; ++iter, ++maskIter)
    {
        uchar maskedWord = *iter & *maskIter;
        if (maskedWord) // the word is not zero
        {
            uchar locWord(maskedWord & ONEBITMASK);
            unsigned int nr3 = DenseVec<BinaryField>::bit_count(maskedWord & (locWord << 1)); // we have counted all (1+X)'s
            unsigned int nr1_3 = DenseVec<BinaryField>::bit_count(locWord); // we have counted all units
            unsigned int nr2 = DenseVec<BinaryField>::bit_count(maskedWord) - (nr3 + nr1_3); // we have counted all X's
            res1 += nr1_3 - nr3;
            res2 += nr2;
            res3 += nr3;

        }
    }
    return res;
        }

SymWeight<2u> DenseVec<F2_F2>::getWeight(void) const
        {
    SymWeight<2u> w;
    std::vector<uchar >::const_iterator iter(m_data.begin());
    std::vector<uchar >::const_iterator iterEnd(m_data.end());
    for (; iter < iterEnd; ++iter)
    {
        if (*iter) // the word is not zero
        {
            uchar locWord(*iter & ONEBITMASK);
            unsigned int nr3 = DenseVec<BinaryField>::bit_count(*iter & (locWord << 1)); // we have counted all (1+X)'s
            unsigned int nr1_3 = DenseVec<BinaryField>::bit_count(locWord); // we have counted all units
            unsigned int nr2 = DenseVec<BinaryField>::bit_count(*iter) - nr3 - nr1_3; // we have counted all X's
            w.add(0, nr1_3);
            w.add(1, nr2);
        }
    }
    return w;
        }

unsigned int DenseVec<F2_F2>::grayImageWeight(void) const
        {
    SymWeight<2u> w(getWeight());
    return w.qty[0] + ( w.qty[1]<<1);
        }

std::ostream& operator<<(std::ostream& os, const DenseVec<F2_F2>& c)
{
    Vec<F2_F2> v(c.getVector());
    os << v << std::endl;
    return os;
}

std::vector<Element<F2_F2> > DenseVec<F2_F2>::getVector(void) const
        {
    std::vector<Element<F2_F2> > res(m_len, 0);
    std::vector<uchar>::const_iterator iterData(m_data.begin());
    std::vector<Element<F2_F2> >::iterator iterVec(res.begin());
    std::vector<Element<F2_F2> >::iterator iterVecEnd(res.end());
    unsigned int wordpos = 0;
    uchar word(*iterData);

    for (; iterVec < iterVecEnd; ++iterVec)
    {
        *iterVec = word & 3;
        ++wordpos;
        if (wordpos >> 5)
        {
            wordpos = 0;
            ++iterData;
            word = *iterData;
        }
        else
        {
            word >>= 2;
        }
    }
    return res;
        }

void DenseVec<F2_F2>::getHeightSeq(std::vector<unsigned int>::iterator iterRes,
        std::vector<unsigned int>::iterator iterResEnd) const
        {
    std::vector<uchar>::const_iterator iterData(m_data.begin());
    unsigned int wordpos = 0;
    uchar word(0);

    for (; iterRes != iterResEnd; ++iterRes)
    {
        if ((wordpos & DenseVec<F2_F2>::BITPOSMASK) == 0) // wordpos % 32 == 0
        {
            wordpos = 0;
            word = *iterData;
            ++iterData;
        }

        if (word & 1)
        {
            *iterRes = 0;
        }
        else if (word & 2)
        {
            *iterRes = 1;
        }
        else
        {
            *iterRes = 2;
        }
        ++wordpos;
        word >>= 2;
    }
        }

